import numbers
import re
from collections.abc import Callable
from collections.abc import Collection
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from collections.abc import Sized
from typing import Any

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import shapely
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from numpy.typing import NDArray
from pandas.api.types import is_array_like
from pandas.api.types import is_dict_like
from pandas.api.types import is_list_like
from pyproj import CRS
from shapely import Geometry
from shapely import box
from shapely import wkb
from shapely import wkt
from shapely.errors import GEOSException
from shapely.geometry import Point
from shapely.geometry import shape
from shapely.ops import unary_union

try:
    import rasterio
    from rasterio import features
except ImportError:
    pass

try:
    from affine import Affine
except ImportError:

    class Affine:
        """Placeholder."""


def crs_to_string(crs: Any) -> str:
    """Extract the string of a CRS-like object."""
    if crs is None:
        return "None"
    crs = pyproj.CRS(crs)
    crs_str = str(crs.to_json_dict()["name"])
    pattern = r"\d{4,5}"
    match = re.search(pattern, crs_str)
    return match.group() if match else crs_str


def to_geoseries(obj: Any, crs: Any | None = None) -> GeoSeries:
    """Convert an object to GeoSeries."""
    if crs is None:
        try:
            crs = obj.crs
        except AttributeError:
            pass

    try:
        if hasattr(obj.index, "values"):
            # pandas objects
            index = obj.index
        else:
            # list
            index = None
    except AttributeError:
        index = None

    try:
        # this works for geodataframe, geoseries and DataFrame with geometry column
        obj = obj.geometry.values
    except AttributeError:
        try:
            # if pandas series
            obj = obj.values
        except AttributeError:
            # geoseries will raise an Exception for non-geometry objects,
            # so we can safely pass here
            pass

    return GeoSeries(obj, index=index, crs=crs)


def to_shapely(obj: Any) -> Geometry:
    """Convert a geometry object or bounding box to a shapely Geometry."""
    if isinstance(obj, Geometry):
        return obj
    if not hasattr(obj, "__iter__"):
        raise TypeError(type(obj))
    try:
        return shapely.union_all(obj.geometry.values)
    except AttributeError:
        pass
    try:
        return Point(*obj)
    except TypeError:
        pass
    try:
        return box(*to_bbox(obj))
    except TypeError:
        pass
    try:
        return shapely.wkt.loads(obj)
    except TypeError:
        pass
    try:
        return shapely.wkb.loads(obj)
    except TypeError:
        pass
    raise TypeError(type(obj), obj)


def to_bbox(
    obj: GeoDataFrame | GeoSeries | Geometry | Collection | Mapping,
) -> tuple[float, float, float, float]:
    """Returns 4-length tuple of bounds if possible, else raises ValueError.

    Args:
        obj: Object to be converted to bounding box. Can be geopandas or shapely
            objects, iterables of exactly four numbers or dictionary like/class
            with a the keys/attributes "minx", "miny", "maxx", "maxy" or
            "xmin", "ymin", "xmax", "ymax".
    """
    if isinstance(obj, (GeoDataFrame, GeoSeries)):
        bounds = tuple(obj.total_bounds)
        assert isinstance(bounds, tuple)
        return bounds
    try:
        bounds = tuple(obj.bounds)
        assert isinstance(bounds, tuple)
        return bounds
    except Exception:
        pass

    def to_int_if_possible(x):
        if isinstance(x, int) or float(x).is_integer():
            return int(x)
        return float(x)

    for attrs in [
        ("minx", "miny", "maxx", "maxy"),
        ("xmin", "ymin", "xmax", "xmax"),
        ("west_longitude", "south_latitude", "east_longitude", "north_latitude"),
    ]:
        try:
            return tuple(to_int_if_possible(obj[attr]) for attr in attrs)
        except Exception:
            pass
        try:
            return tuple(to_int_if_possible(getattr(obj, attr)) for attr in attrs)
        except Exception:
            pass

    if hasattr(obj, "geometry"):
        try:
            return tuple(GeoSeries(obj["geometry"]).total_bounds)  # type: ignore [index]
        except Exception:
            return tuple(GeoSeries(obj.geometry).total_bounds)

    if (
        hasattr(obj, "__len__")
        and len(obj) == 4
        and all(isinstance(x, numbers.Number) for x in obj)
    ):
        return tuple(obj)

    try:
        of_length = f" of length {len(obj)}"
    except TypeError:
        of_length = ""
    raise TypeError(f"Cannot convert type {obj.__class__.__name__}{of_length} to bbox")


def from_4326(lon: float, lat: float, crs=25833) -> tuple[float, float]:
    """Get utm 33 N coordinates from lonlat (4326)."""
    transformer = pyproj.Transformer.from_crs(
        "EPSG:4326", f"EPSG:{crs}", always_xy=True
    )
    return transformer.transform(lon, lat)


def to_4326(lon: float, lat: float, crs=25833) -> tuple[float, float]:
    """Get degree coordinates  33 N coordinates from lonlat (4326)."""
    transformer = pyproj.Transformer.from_crs(
        f"EPSG:{crs}", "EPSG:4326", always_xy=True
    )
    return transformer.transform(lon, lat)


def coordinate_array(
    gdf: GeoDataFrame | GeoSeries,
    strict: bool = False,
    include_z: bool = False,
) -> NDArray[np.float64]:
    """Creates a 2d ndarray of coordinates from point geometries.

    Args:
        gdf: GeoDataFrame or GeoSeries of point geometries.
        strict: If False (default), geometries without coordinates
            are given the value None.
        include_z: Whether to include z-coordinates. Defaults to False.

    Returns:
        np.ndarray of np.ndarrays of coordinates.

    Examples:
    ---------
    >>> import sgis as sg
    >>> points = sg.to_gdf(
    ...     [
    ...         (0, 1),
    ...         (1, 0),
    ...         (1, 1),
    ...         (0, 0),
    ...         (0.5, 0.5),
    ...         (0.5, 0.25),
    ...         (0.25, 0.25),
    ...         (0.75, 0.75),
    ...         (0.25, 0.75),
    ...         (0.75, 0.25),
    ...     ]
    ... )
    >>> points
                    geometry
    0  POINT (0.59376 0.92577)
    1  POINT (0.34075 0.91650)
    2  POINT (0.74841 0.10627)
    3  POINT (0.00966 0.87868)
    4  POINT (0.38046 0.87879)
    >>> sg.coordinate_array(points)
    array([[0.59376221, 0.92577159],
        [0.34074678, 0.91650446],
        [0.74840912, 0.10626954],
        [0.00965935, 0.87867915],
        [0.38045827, 0.87878816]])
    >>> sg.coordinate_array(points.geometry)
    array([[0.59376221, 0.92577159],
        [0.34074678, 0.91650446],
        [0.74840912, 0.10626954],
        [0.00965935, 0.87867915],
        [0.38045827, 0.87878816]])
    """
    try:
        return shapely.get_coordinates(gdf.geometry.values, include_z=include_z)
    except AttributeError:
        return shapely.get_coordinates(gdf, include_z=include_z)


def to_gdf(
    obj: (
        Geometry
        | str
        | bytes
        | list
        | tuple
        | dict
        | GeoSeries
        | pd.Series
        | pd.DataFrame
        | Iterator
    ),
    crs: str | tuple[str] | None = None,
    geometry: str | tuple[str] | int | None = None,
    **kwargs,
) -> GeoDataFrame:
    """Converts geometry-like objects to a GeoDataFrame.

    Constructs a GeoDataFrame from any geometry-like object
    (coordinates, wkt, wkb, dict, string) or any interable of such objects.

    NOTE: The function is meant for convenience in testing and exploring,
    not for production code since it introduces unnecessary overhead
    and potential errors.

    If obj is a DataFrame or dictionary, geometries can be in one column/key or 2-3
    if coordiantes are in x and x (and z) columns. The column/key "geometry" is used
    by default if it exists. The index and other columns/keys are preserved.

    Args:
        obj: the object to be converted to a GeoDataFrame.
        crs: if None (default), it uses the crs of the GeoSeries if GeoSeries
            is the input type. Otherwise, no crs is used.
        geometry: Name of column(s) or key(s) in obj with geometry-like values.
            If not specified, the key/column 'geometry' will be used if it
            exists. If multiple columns, can be given as e.g. "xyz" or ["x", "y"].
        **kwargs: additional keyword arguments taken by the GeoDataFrame constructor.

    Returns:
        A GeoDataFrame with one column, the geometry column.

    Examples:
    ---------
    >>> import sgis as sg
    >>> coords = (10, 60)
    >>> sg.to_gdf(coords, crs=4326)
                        geometry
    0  POINT (10.00000 60.00000)

    From wkt.

    >>> wkt = "POINT (10 60)"
    >>> sg.to_gdf(wkt, crs=4326)
                        geometry
    0  POINT (10.00000 60.00000)

    From DataFrame with x, y (optionally z) coordinate columns. Index and
    columns are preserved.

    >>> df = pd.DataFrame({"x": [10, 11], "y": [60, 59]}, index=[1,3])
        x   y
    1  10  60
    3  11  59
    >>> gdf = sg.to_gdf(df, geometry=["x", "y"], crs=4326)
    >>> gdf
        x   y                   geometry
    1  10  60  POINT (10.00000 60.00000)
    3  11  59  POINT (11.00000 59.00000)

    For DataFrame/dict with a geometry-like column named "geometry". If the column has
    another name, it must be set with the geometry parameter.

    >>> df = pd.DataFrame({"col": [1, 2], "geometry": ["point (10 60)", (11, 59)]})
    >>> df
       col       geometry
    0    1  point (10 60)
    1    2       (11, 59)
    >>> gdf = sg.to_gdf(df, crs=4326)
    >>> gdf
       col                   geometry
    0    1  POINT (10.00000 60.00000)
    1    2  POINT (11.00000 59.00000)

    From Series.

    >>> series = Series({1: (10, 60), 3: (11, 59)})
    >>> sg.to_gdf(series)
                        geometry
    1  POINT (10.00000 60.00000)
    3  POINT (11.00000 59.00000)

    Multiple coordinates will be converted to points, unless a line or polygon geometry
    is constructed beforehand.

    >>> coordslist = [(10, 60), (11, 59)]
    >>> sg.to_gdf(coordslist, crs=4326)
                        geometry
    0  POINT (10.00000 60.00000)
    1  POINT (11.00000 59.00000)

    >>> from shapely.geometry import LineString
    >>> sg.to_gdf(LineString(coordslist), crs=4326)
                                                geometry
    0  LINESTRING (10.00000 60.00000, 11.00000 59.00000)

    From 2 or 3 dimensional array.

    >>> arr = np.random.randint(100, size=(5, 3))
    >>> sg.to_gdf(arr)
                             geometry
    0  POINT Z (82.000 88.000 82.000)
    1  POINT Z (70.000 92.000 20.000)
    2   POINT Z (91.000 34.000 3.000)
    3   POINT Z (1.000 50.000 77.000)
    4  POINT Z (58.000 49.000 46.000)
    """
    if isinstance(obj, GeoDataFrame):
        if not crs:
            return obj
        if not obj.crs:
            return obj.set_crs(crs)
        return obj.to_crs(crs)

    if obj is None:
        raise TypeError("Cannot convert NoneType to GeoDataFrame.")

    if isinstance(obj, GeoSeries):
        geom_col = geometry or "geometry"
        return _geoseries_to_gdf(obj, geom_col, crs, **kwargs)

    if crs is None:
        try:
            crs = obj.crs  # type: ignore
        except AttributeError:
            try:
                matches = re.search(r"SRID=(\d+);", obj)  # type: ignore
            except TypeError:
                try:
                    matches = re.search(r"SRID=(\d+);", obj[0])  # type: ignore
                except Exception:
                    pass
            try:
                crs = CRS(
                    int(
                        "".join(
                            x
                            for x in matches.group(0)  # type:ignore
                            if x.isnumeric()
                        )
                    )
                )
            except Exception:
                pass

    if is_array_like(geometry) and len(geometry) == len(obj):  # type: ignore
        geometry = GeoSeries(
            _make_one_shapely_geom(g) for g in geometry if g is not None  # type: ignore
        )
        return GeoDataFrame(obj, geometry=geometry, crs=crs, **kwargs)

    geom_col: str = _find_geometry_column(obj, geometry)  # type: ignore[no-redef]
    index = kwargs.pop("index", None)

    # get done with iterators that would get consumed by 'all' later
    if isinstance(obj, Iterator) and not isinstance(obj, Sized):
        obj = list(obj)

    if hasattr(obj, "__len__") and not len(obj):
        return GeoDataFrame({"geometry": []}, crs=crs)

    crs = crs or get_crs_from_dict(obj)

    if not is_dict_like(obj):
        if is_bbox_like(obj):
            obj = GeoSeries(shapely.box(*obj), index=index)
            return GeoDataFrame({geom_col: obj}, geometry=geom_col, crs=crs, **kwargs)
        elif all(hasattr(obj, attr) for attr in ["minx", "miny", "maxx", "maxy"]):
            obj = GeoSeries(shapely.box(*to_bbox(obj)), index=index)
            return GeoDataFrame({geom_col: obj}, geometry=geom_col, crs=crs, **kwargs)
        if is_nested_geojson(obj):
            # crs = crs or get_crs_from_dict(obj)
            obj = pd.concat(
                (GeoSeries(_from_json(g)) for g in obj if g is not None),  # type: ignore
                ignore_index=True,
            )
            if index is not None:
                obj.index = index
            return GeoDataFrame({geom_col: obj}, geometry=geom_col, crs=crs, **kwargs)
        else:
            if isinstance(obj, str):
                if (
                    obj.replace(" ", "")
                    .replace(",", "")
                    .replace(".", "")
                    .replace("(", "")
                    .replace(")", "")
                    .isnumeric()
                ):
                    # string of coordinates
                    obj = eval(obj)
                    obj = GeoSeries([_make_one_shapely_geom(obj)], index=index)
                else:
                    # wkt
                    obj = GeoSeries([_make_one_shapely_geom(obj)], index=index)
            else:
                # list etc.
                obj = GeoSeries(_make_shapely_geoms(obj), index=index)
            return GeoDataFrame(
                {geom_col: obj}, geometry=geom_col, index=index, crs=crs, **kwargs
            )

    # now we have dict, Series or DataFrame

    obj = obj.copy()  # type: ignore [union-attr]

    # preserve Series/DataFrame index
    index = obj.index if hasattr(obj, "index") and index is None else index

    if geom_col in obj.keys():
        if isinstance(obj, pd.DataFrame):
            notna = obj[geom_col].notna()
            obj.loc[notna, geom_col] = list(
                _make_shapely_geoms(obj.loc[notna, geom_col])
            )
            obj[geom_col] = GeoSeries(obj[geom_col])
            return GeoDataFrame(obj, geometry=geom_col, crs=crs, **kwargs)
        if isinstance(obj[geom_col], Geometry):
            return GeoDataFrame(
                dict(obj), geometry=geom_col, crs=crs, index=[0], **kwargs
            )
        if not hasattr(obj[geom_col], "__iter__") or len(obj[geom_col]) == 1:
            obj[geom_col] = _make_shapely_geoms(obj[geom_col])
            return GeoDataFrame(
                dict(obj), geometry=geom_col, crs=crs, index=index, **kwargs
            )
        obj[geom_col] = GeoSeries(_make_shapely_geoms(obj[geom_col]), index=index)
        return GeoDataFrame(dict(obj), geometry=geom_col, crs=crs, **kwargs)

    if geometry and all(g in obj for g in geometry):  # type: ignore [union-attr]
        obj[geom_col] = _geoseries_from_xyz(obj, geometry, index=index)
        return GeoDataFrame(obj, geometry=geom_col, crs=crs, **kwargs)

    if len(obj.keys()) == 1:
        key = next(iter(obj.keys()))
        if isinstance(obj, dict):
            geoseries = GeoSeries(
                _make_shapely_geoms(next(iter(obj.values()))), index=index
            )
        elif isinstance(obj, pd.Series):
            geoseries = GeoSeries(_make_shapely_geoms(obj), index=index)
        else:
            geoseries = GeoSeries(_make_shapely_geoms(obj.iloc[:, 0]), index=index)
        return GeoDataFrame({key: geoseries}, geometry=key, crs=crs, **kwargs)

    if (geometry and geom_col not in obj) or isinstance(obj, pd.DataFrame):
        raise ValueError("Cannot find geometry column(s)", geometry)

    # geojson, __geo_interface__
    if (
        isinstance(obj, dict)
        and sum(key in obj for key in ["type", "coordinates", "features"]) >= 2
    ):
        if "geometry" in obj:
            geometry = "geometry"

        # crs = crs or get_crs_from_dict(obj)
        obj = GeoSeries(_from_json(obj), index=index)
        return GeoDataFrame({geom_col: obj}, geometry=geom_col, crs=crs, **kwargs)

    try:
        geoseries = _series_like_to_geoseries(obj, index=index)
    except ValueError:
        geoseries = _series_like_to_geoseries(obj.dropna(), index=obj.dropna().index)
    return GeoDataFrame(geometry=geoseries, crs=crs, **kwargs)


def _array_to_geojson(
    array: np.ndarray, transform: Affine
) -> list[tuple[dict, Geometry]]:
    try:
        return [
            (value, shape(geom))
            for geom, value in features.shapes(array, transform=transform)
        ]
    except ValueError:
        array = array.astype(np.float32)
        return [
            (value, shape(geom))
            for geom, value in features.shapes(array, transform=transform)
        ]


def get_transform_from_bounds(
    obj: GeoDataFrame | GeoSeries | Geometry | tuple, shape: tuple[float, ...]
) -> Affine:
    minx, miny, maxx, maxy = to_bbox(obj)
    if len(shape) == 2:
        width, height = shape
    elif len(shape) == 3:
        _, width, height = shape
    else:
        raise ValueError
    return rasterio.transform.from_bounds(minx, miny, maxx, maxy, width, height)


def _make_shapely_geoms(obj: Any) -> Geometry | Any:
    if _is_one_geometry(obj):
        return _make_one_shapely_geom(obj)
    if isinstance(obj, dict) and "coordinates" in obj:
        return _from_json(obj)
    return (_make_one_shapely_geom(g) for g in obj)


def is_bbox_like(obj: Any) -> bool:
    """Returns True if the object is an iterable of 4 numbers.

    Args:
        obj: Any object.
    """
    if (
        hasattr(obj, "__len__")
        and len(obj) == 4
        and hasattr(obj, "__iter__")
        and all(isinstance(x, numbers.Number) for x in obj)
    ):
        return True
    return False


def is_nested_geojson(obj: Any) -> bool:
    """Returns True if the object is an iterable of all dicts.

    Args:
        obj: Any object.
    """
    if hasattr(obj, "__iter__") and all(isinstance(g, dict) for g in obj):
        return True
    return False


def get_crs_from_dict(obj: Any) -> CRS | None | Any:
    """Try to extract the 'crs' attribute of the object or an object in the object."""
    if not hasattr(obj, "__iter__") or (
        not is_dict_like(obj) and not is_dict_like(obj[0])
    ):
        return None

    if not is_dict_like(obj) and is_dict_like(obj[0]):
        crss = list({get_crs_from_dict(g) for g in obj})
        return crss[0] if len(crss) == 1 else None

    if "properties" in obj:
        return get_crs_from_dict(obj["properties"])

    if "crs" in obj:
        obj = obj["crs"]
        while is_dict_like(obj):
            if "properties" in obj:
                obj = obj["properties"]
            elif "name" in obj:
                obj = obj["name"]
            else:
                return None
        return obj

    return None


def _from_json(obj: dict | list[dict]) -> Geometry:
    if not isinstance(obj, dict) and isinstance(obj[0], dict):
        return [_from_json(g) for g in obj]
    if "geometry" in obj:
        return _from_json(obj["geometry"])
    if "features" in obj:
        return _from_json(obj["features"])
    coords = obj["coordinates"]
    constructor: Callable = eval("shapely.geometry." + obj.get("type", Point))
    try:
        return constructor(coords)
    except TypeError:
        while len(coords) == 1:
            coords = coords[0]
        return constructor(coords)


def _series_like_to_geoseries(obj: Iterable, index: Iterable) -> GeoSeries:
    if index is None:
        index = obj.keys()
    if isinstance(obj, dict):
        return GeoSeries(_make_shapely_geoms(list(obj.values())), index=index)
    else:
        return GeoSeries(_make_shapely_geoms(obj.values), index=index)


def _geoseries_to_gdf(
    obj: GeoSeries, geometry: str | GeoSeries | Iterable, crs: CRS | Any, **kwargs
) -> GeoDataFrame:
    if not crs:
        crs = obj.crs
    else:
        obj = obj.to_crs(crs) if obj.crs else obj.set_crs(crs)

    return GeoDataFrame({geometry: obj}, geometry=geometry, crs=crs, **kwargs)


def _find_geometry_column(obj: Any, geometry: GeoSeries | Iterable | None) -> str:
    if geometry is None:
        return "geometry"

    # dict key
    if not is_list_like(geometry) and geometry in obj:
        return geometry

    # nested dict key
    if len(geometry) == 1 and geometry[0] in obj:
        return geometry[0]

    if len(geometry) in {2, 3}:
        return "geometry"

    raise ValueError(
        "geometry should be a geometry column or x, y (z) coordinate columns."
    )


def _geoseries_from_xyz(
    obj: Any,
    geometry: Iterable[float, float] | Iterable[float, float, float],
    index: Iterable | None,
) -> GeoSeries:
    """Make geoseries from the geometry column or columns (x y (z))."""
    if len(geometry) == 2:
        x, y = geometry
        z = None

    elif len(geometry) == 3:
        x, y, z = geometry
        z = obj[z]

    else:
        raise ValueError(
            "geometry should be a geometry column or x, y (z) coordinate columns."
        )

    return gpd.GeoSeries.from_xy(x=obj[x], y=obj[y], z=z, index=index)


def _is_one_geometry(obj: Any) -> bool:
    if (
        isinstance(obj, (str, bytes, Geometry))  # type: ignore [unreachable]
        or all(isinstance(i, numbers.Number) for i in obj)
        or not hasattr(obj, "__iter__")
    ):
        return True
    return False  # type: ignore [unreachable]


def _make_one_shapely_geom(obj: Any) -> Geometry:
    """Create shapely geometry from wkt, wkb or coordinate tuple.

    Works recursively if the object is a nested iterable.
    """
    if isinstance(obj, str):
        try:
            return wkt.loads(obj)
        except GEOSException:
            if obj.startswith("geography"):
                matches = re.search(r"SRID=(\d+);", obj)
                srid = matches.group(0)
                _, _wkt = obj.split(srid)
                return wkt.loads(_wkt)

    if isinstance(obj, bytes):
        return wkb.loads(obj)

    if isinstance(obj, Geometry):
        return obj

    if not hasattr(obj, "__iter__"):
        raise ValueError(
            f"Couldn't create shapely geometry from {obj} of type {type(obj)}"
        )

    if isinstance(obj, GeoSeries):
        raise TypeError(
            "to_gdf doesn't accept iterable of GeoSeries. Instead use: "
            "pd.concat(to_gdf(obj) for obj in geoseries_iterable)"
        )

    if not any(isinstance(g, numbers.Number) for g in obj):
        # we're likely dealing with a nested iterable, so let's
        # recursively dig down to the coords/wkt/wkb
        if len(obj) == 2 or len(obj) == 3:
            try:
                obj = [float(g) for g in obj]
                return Point(obj)
            except Exception:
                pass
        return unary_union([_make_one_shapely_geom(g) for g in obj])

    elif len(obj) == 2 or len(obj) == 3:
        return Point(obj)

    elif len(obj) == 4:
        return box(*obj)
    else:
        raise ValueError(
            "If 'geom' is an iterable, each item should consist of "
            "wkt, wkb or (x, y (z) or bbox). Got ",
            obj,
        )



from collections.abc import Callable
from collections.abc import Sequence

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely import Geometry

from ..helpers import get_non_numpy_func_name
from ..helpers import get_numpy_func


def _prepare_zonal(
    polygons: gpd.GeoDataFrame,
    aggfunc: str | Callable | Sequence[Callable | str],
) -> tuple[gpd.GeoDataFrame, list[Callable], list[str]]:
    polygons = polygons.reset_index(drop=True)[["geometry"]]

    if isinstance(aggfunc, str) or callable(aggfunc):
        aggfunc = [aggfunc]

    aggfunc = [f if callable(f) else get_numpy_func(f) for f in aggfunc]

    func_names = [get_non_numpy_func_name(f) for f in aggfunc]

    return polygons, aggfunc, func_names


def _make_geometry_iterrows(gdf: gpd.GeoDataFrame) -> list[tuple[int, Geometry]]:
    """Because pandas iterrows returns Series, not GeoSeries."""
    return list(gdf.geometry.items())


def _no_overlap_df(func_names: list[str], i: int, date: str) -> pd.DataFrame:
    df = pd.DataFrame(columns=func_names, index=[i])
    df["date"] = date
    df["_no_overlap"] = 1
    return df


def _aggregate(
    array: np.ndarray,
    array_func: Callable,
    aggfunc: Callable,
    func_names: list[str],
    date: str,
    i: int,
) -> pd.DataFrame:
    if array_func:
        array = array_func(array)
    # flat_array = array.astype(np.float64).flatten()
    flat_array = array.flatten()
    no_nans = flat_array[~np.isnan(flat_array)]
    data = {}
    for f, name in zip(aggfunc, func_names, strict=True):
        num = f(no_nans)
        data[name] = num
    df = pd.DataFrame(data, index=[i])
    df["date"] = date
    df["_no_overlap"] = 0
    return df


def _zonal_post(
    aggregated: list[pd.DataFrame],
    polygons: gpd.GeoDataFrame,
    idx_mapper: dict | pd.Series,
    idx_name: str,
    dropna: bool,
) -> pd.DataFrame:
    out = gpd.GeoDataFrame(
        pd.concat(aggregated), geometry=polygons.geometry.values, crs=polygons.crs
    ).sort_index()

    out.index = out.index.map(idx_mapper)
    out.index.name = idx_name

    if dropna:
        out = out.loc[~out.drop(columns="geometry").isna().all(axis=1)]
        return out.loc[out["_no_overlap"] != 1].drop(columns="_no_overlap")
    else:
        return out.drop(columns="_no_overlap")



"""Function to check whether we are in Dapla without importing Dapla.

Function from: From https://github.com/statisticsnorway/ssb-altinn-python/blob/main/src/altinn/utils.py
"""

import os


def is_dapla() -> bool:
    """Simply checks if an os environment variable contains the text 'dapla'."""
    return any("dapla" in key.lower() for key in os.environ)



import re
import warnings
from typing import Any

import numpy as np
import pandas as pd
import shapely
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from geopandas.array import GeometryArray
from numpy.typing import NDArray
from shapely import extract_unique_points
from shapely import force_2d
from shapely import get_coordinates
from shapely import get_exterior_ring
from shapely import get_parts
from shapely import linearrings
from shapely import linestrings
from shapely import make_valid
from shapely import polygons
from shapely.errors import GEOSException
from shapely.geometry import LinearRing
from shapely.geometry import LineString
from shapely.geometry import Point

from ..networkanalysis.closing_network_holes import get_angle
from .buffer_dissolve_explode import buff
from .buffer_dissolve_explode import dissexp
from .buffer_dissolve_explode import dissexp_by_cluster
from .conversion import coordinate_array
from .conversion import to_geoseries
from .duplicates import get_intersections
from .duplicates import update_geometries
from .general import clean_geoms
from .general import sort_large_first
from .general import sort_small_first
from .general import to_lines
from .geometry_types import make_all_singlepart
from .overlay import clean_overlay
from .polygon_operations import close_all_holes
from .polygon_operations import eliminate_by_longest
from .polygon_operations import get_cluster_mapper
from .polygon_operations import get_gaps
from .polygons_as_rings import PolygonsAsRings
from .sfilter import sfilter

warnings.simplefilter(action="ignore", category=UserWarning)
warnings.simplefilter(action="ignore", category=RuntimeWarning)


PRECISION = 1e-4
BUFFER_RES = 50


def coverage_clean(
    gdf: GeoDataFrame,
    tolerance: int | float,
    mask=None,
    *,
    duplicate_action: str = "fix",
    grid_sizes: tuple[None | int] = (None,),
    logger=None,
) -> GeoDataFrame:
    """Fix thin gaps, holes, slivers and double surfaces.

    Rules:
    - Holes (interiors) thinner than the tolerance are closed.
    - Gaps between polygons are filled if thinner than the tolerance.
    - Sliver polygons thinner than the tolerance are eliminated
    into the neighbor polygon with the longest shared border.
    - Double surfaces thinner than the tolerance are eliminated.
    If duplicate_action is "fix", thicker double surfaces will
    be updated from top to bottom of the GeoDataFrame's rows.
    - Line and point geometries are removed.
    - MultiPolygons are exploded to Polygons.
    - Index is reset.

    Args:
        gdf: GeoDataFrame to be cleaned.
        tolerance: distance (usually meters) used as the minimum thickness
            for polygons to be eliminated. Any gap, hole, sliver or double
            surface that are empty after a negative buffer of tolerance / 2
            are eliminated into the neighbor with the longest shared border.
        mask: Unused.
        duplicate_action: Either "fix", "error" or "ignore".
            If "fix" (default), double surfaces thicker than the
            tolerance will be updated from top to bottom (function update_geometries)
            and then dissolved into the neighbor polygon with the longest shared border.
            If "error", an Exception is raised if there are any double surfaces thicker
            than the tolerance. If "ignore", double surfaces are kept as is.
        grid_sizes: One or more grid_sizes used in overlay and dissolve operations that
            might raise a GEOSException. Defaults to (None,), meaning no grid_sizes.
        logger: Optional.

    Returns:
        A GeoDataFrame with cleaned polygons.

    """
    if not len(gdf):
        return gdf

    _cleaning_checks(gdf, tolerance, duplicate_action)

    # if not gdf.index.is_unique:
    #     gdf = gdf.reset_index(drop=True)

    # gdf = make_all_singlepart(gdf).loc[
    #     lambda x: x.geom_type.isin(["Polygon", "MultiPolygon"])
    # ]

    # gdf = safe_simplify(gdf, PRECISION)

    gdf = (
        clean_geoms(gdf)
        .pipe(make_all_singlepart, ignore_index=True)
        .loc[lambda x: x.geom_type.isin(["Polygon", "MultiPolygon"])]
    )

    try:
        gaps = get_gaps(gdf, include_interiors=True)
    except GEOSException:
        for i, grid_size in enumerate(grid_sizes):
            try:
                gaps = get_gaps(gdf, include_interiors=True, grid_size=grid_size)
                if grid_size:
                    # in order to not get more gaps
                    gaps.geometry = gaps.buffer(grid_size)
                break
            except GEOSException as e:
                if i == len(grid_sizes) - 1:
                    explore_geosexception(e, gdf, logger=logger)
                    raise e

    gaps["_was_gap"] = 1

    if duplicate_action == "ignore":
        double = GeoDataFrame({"geometry": []}, crs=gdf.crs)
        double["_double_idx"] = None
    else:
        double = get_intersections(gdf)
        double["_double_idx"] = range(len(double))

    gdf, slivers = split_out_slivers(gdf, tolerance)

    gdf["_poly_idx"] = range(len(gdf))

    thin_gaps_and_double = pd.concat([gaps, double]).loc[
        lambda x: x.buffer(-tolerance / 2).is_empty
    ]

    all_are_thin = double["_double_idx"].isin(thin_gaps_and_double["_double_idx"]).all()

    if not all_are_thin and duplicate_action == "fix":
        gdf, thin_gaps_and_double, slivers = _properly_fix_duplicates(
            gdf, double, slivers, thin_gaps_and_double, tolerance
        )

    elif not all_are_thin and duplicate_action == "error":
        raise ValueError("Large double surfaces.")

    to_eliminate = pd.concat([thin_gaps_and_double, slivers], ignore_index=True)
    to_eliminate = safe_simplify(to_eliminate, PRECISION)

    to_eliminate = to_eliminate.loc[lambda x: ~x.buffer(-PRECISION / 10).is_empty]

    to_eliminate = try_for_grid_size(
        split_by_neighbors,
        grid_sizes=grid_sizes,
        args=(to_eliminate, gdf),
        kwargs=dict(tolerance=tolerance),
    )

    to_eliminate["_eliminate_idx"] = range(len(to_eliminate))

    to_eliminate["_cluster"] = get_cluster_mapper(to_eliminate.buffer(PRECISION))

    gdf_geoms_idx = gdf[["_poly_idx", "geometry"]]

    poly_idx_mapper: pd.Series = (
        clean_overlay(
            buff(
                to_eliminate[["_eliminate_idx", "geometry"]],
                tolerance,
                resolution=BUFFER_RES,
            ),
            gdf_geoms_idx,
            geom_type="polygon",
        )
        .pipe(sort_large_first)
        .drop_duplicates("_eliminate_idx")
        .set_index("_eliminate_idx")["_poly_idx"]
    )

    to_eliminate["_poly_idx"] = to_eliminate["_eliminate_idx"].map(poly_idx_mapper)
    isolated = to_eliminate[lambda x: x["_poly_idx"].isna()]
    intersecting = to_eliminate[lambda x: x["_poly_idx"].notna()]

    for i, grid_size in enumerate(grid_sizes):
        try:
            without_double = update_geometries(
                intersecting, geom_type="polygon", grid_size=grid_size
            ).drop(columns=["_eliminate_idx", "_double_idx"])
            break
        except GEOSException as e:
            if i == len(grid_sizes) - 1:
                explore_geosexception(e, gdf, intersecting, isolated, logger=logger)
                raise e

    not_really_isolated = isolated[["geometry", "_eliminate_idx", "_cluster"]].merge(
        without_double.drop(columns=["geometry"]),
        on="_cluster",
        how="inner",
    )

    really_isolated = isolated.loc[
        lambda x: ~x["_eliminate_idx"].isin(not_really_isolated["_eliminate_idx"])
    ]

    is_gap = really_isolated["_was_gap"] == 1
    isolated_gaps = really_isolated.loc[is_gap, ["geometry"]].sjoin_nearest(
        gdf, max_distance=PRECISION
    )
    really_isolated = really_isolated[~is_gap]

    really_isolated["_poly_idx"] = (
        really_isolated["_cluster"] + gdf["_poly_idx"].max() + 1
    )

    for i, grid_size in enumerate(grid_sizes):
        try:
            cleaned = (
                dissexp(
                    pd.concat(
                        [
                            gdf,
                            without_double,
                            not_really_isolated,
                            really_isolated,
                            isolated_gaps,
                        ]
                    ).drop(
                        columns=[
                            "_cluster",
                            "_was_gap",
                            "_eliminate_idx",
                            "index_right",
                            "_double_idx",
                        ],
                        errors="ignore",
                    ),
                    by="_poly_idx",
                    aggfunc="first",
                    dropna=True,
                    grid_size=grid_size,
                )
                .sort_index()
                .reset_index(drop=True)
            )
            break
        except GEOSException as e:
            if i == len(grid_sizes) - 1:
                explore_geosexception(
                    e, gdf, without_double, isolated, really_isolated, logger=logger
                )
                raise e

    cleaned_area_sum = cleaned.area.sum()

    for i, grid_size in enumerate(grid_sizes):
        try:
            cleaned = clean_overlay(
                gdf.drop(columns="_poly_idx"),
                cleaned,
                how="update",
                geom_type="polygon",
            )
            break
        except GEOSException as e:
            if i == len(grid_sizes) - 1:
                explore_geosexception(
                    e,
                    gdf,
                    cleaned,
                    without_double,
                    isolated,
                    really_isolated,
                    logger=logger,
                )
                raise e
            try:
                cleaned = update_geometries(
                    cleaned, geom_type="polygon", grid_size=grid_size
                )
            except GEOSException:
                pass

    # if logger and cleaned_area_sum > cleaned.area.sum() + 1:
    #     print("\ncleaned.area.sum() diff", cleaned_area_sum - cleaned.area.sum())
    #     logger.debug("cleaned.area.sum() diff", cleaned_area_sum - cleaned.area.sum())

    cleaned = sort_large_first(cleaned)

    for i, grid_size in enumerate(grid_sizes):
        try:
            cleaned = update_geometries(
                cleaned, geom_type="polygon", grid_size=grid_size
            )
            break
        except GEOSException as e:
            cleaned.geometry = shapely.simplify(
                cleaned.geometry, PRECISION * (10 * i + 1)
            )
            if i == len(grid_sizes) - 1:
                explore_geosexception(
                    e,
                    gdf,
                    cleaned,
                    without_double,
                    isolated,
                    really_isolated,
                    logger=logger,
                )
                raise e

    cleaned = safe_simplify(cleaned, PRECISION)
    cleaned.geometry = shapely.make_valid(cleaned.geometry)

    return cleaned


def safe_simplify(gdf, tolerance: float | int) -> GeoDataFrame:
    """Simplify only if the resulting area is no more than 1 percent larger.

    Because simplifying can result in holes being filled.
    """
    length_then = gdf.length
    copied = gdf.copy()
    copied.geometry = shapely.make_valid(
        shapely.simplify(copied.geometry.values, tolerance=tolerance)
    )
    copied.loc[copied.area > length_then * 1.01, copied._geometry_column_name] = (
        gdf.loc[copied.area > length_then * 1.01, copied._geometry_column_name]
    )

    return copied


def remove_spikes(gdf: GeoDataFrame, tolerance: int | float) -> GeoDataFrame:
    """Remove thin spikes in polygons.

    Note that this function might be slow. Should only be used if nessecary.

    Args:
        gdf: GeoDataFrame of polygons
        tolerance: distance (usually meters) used as the minimum thickness
            for polygons to be eliminated. Any spike thinner than the tolerance
            will be removed.

    Returns:
        A GeoDataFrame of polygons without spikes thinner.
    """
    gdf.geometry = (
        PolygonsAsRings(gdf.geometry)
        .apply_numpy_func(_remove_spikes, args=(tolerance,))
        .to_numpy()
    )
    return gdf


# def remove_spikes(gdf: GeoDataFrame, tolerance: int | float) -> GeoDataFrame:
#     return clean_overlay(
#         gdf, gdf[["geometry"]], how="intersection", grid_size=tolerance
#     )


def _remove_spikes(
    geoms: NDArray[LinearRing], tolerance: int | float
) -> NDArray[LinearRing]:
    if not len(geoms):
        return geoms
    geoms = to_geoseries(geoms).reset_index(drop=True)

    points = (
        extract_unique_points(geoms).explode(index_parts=False).to_frame("geometry")
    )

    points = get_angle_between_indexed_points(points)

    def to_buffered_rings_without_spikes(x):
        polys = GeoSeries(make_valid(polygons(get_exterior_ring(x))))

        return (
            polys.buffer(-tolerance, resolution=BUFFER_RES)
            .explode(index_parts=False)
            .pipe(close_all_holes)
            .pipe(get_exterior_ring)
            .buffer(tolerance * 10)
        )

    buffered = to_buffered_rings_without_spikes(
        geoms.buffer(tolerance / 2, resolution=BUFFER_RES)
    )

    points_without_spikes = (
        extract_unique_points(geoms)
        .explode(index_parts=False)
        .loc[lambda x: x.index.isin(sfilter(x, buffered).index)]
    )

    # linearrings require at least 4 coordinate pairs, or three unique
    points_without_spikes = points_without_spikes.loc[
        lambda x: x.groupby(level=0).size() >= 3
    ]

    # need an index from 0 to n-1 in 'linearrings'
    to_int_index = {
        ring_idx: i
        for i, ring_idx in enumerate(sorted(set(points_without_spikes.index)))
    }
    int_indices = points_without_spikes.index.map(to_int_index)

    as_lines = pd.Series(
        linearrings(
            get_coordinates(points_without_spikes.geometry.values),
            indices=int_indices,
        ),
        index=points_without_spikes.index.unique(),
    )

    # the missing polygons are thin and/or spiky. Let's remove them
    missing = geoms.loc[~geoms.index.isin(as_lines.index)]

    missing = pd.Series(
        [None] * len(missing),
        index=missing.index.values,
    )

    return pd.concat([as_lines, missing]).sort_index()


def get_angle_between_indexed_points(point_df: GeoDataFrame) -> GeoDataFrame:
    """Get angle difference between the two lines."""
    point_df["next"] = point_df.groupby(level=0)["geometry"].shift(-1)

    notna = point_df["next"].notna()

    this = coordinate_array(point_df.loc[notna, "geometry"].values)
    next_ = coordinate_array(point_df.loc[notna, "next"].values)

    point_df.loc[notna, "angle"] = get_angle(this, next_)
    point_df["prev_angle"] = point_df.groupby(level=0)["angle"].shift(1)

    point_df["angle_diff"] = point_df["angle"] - point_df["prev_angle"]

    return point_df


def _properly_fix_duplicates(gdf, double, slivers, thin_gaps_and_double, tolerance):
    gdf = _dissolve_thick_double_and_update(gdf, double, thin_gaps_and_double)
    gdf, more_slivers = split_out_slivers(gdf, tolerance)
    slivers = pd.concat([slivers, more_slivers], ignore_index=True)
    gaps = get_gaps(gdf, include_interiors=True)
    gaps["_was_gap"] = 1
    assert "_double_idx" not in gaps
    double = get_intersections(gdf)
    double["_double_idx"] = range(len(double))
    thin_gaps_and_double = pd.concat([gaps, double], ignore_index=True).loc[
        lambda x: x.buffer(-tolerance / 2).is_empty
    ]

    return gdf, thin_gaps_and_double, slivers


def _dissolve_thick_double_and_update(gdf, double, thin_double):
    large = (
        double.loc[~double["_double_idx"].isin(thin_double["_double_idx"])]
        .drop(columns="_double_idx")
        .pipe(sort_small_first)
        # .sort_values("_poly_idx")
        .pipe(update_geometries, geom_type="polygon")
    )
    return (
        clean_overlay(gdf, large, how="update", geom_type="polygon").pipe(
            sort_small_first
        )
        # .sort_values("_poly_idx")
        .pipe(update_geometries, geom_type="polygon")
    )


def _cleaning_checks(gdf, tolerance, duplicate_action):  # , spike_action):
    if not len(gdf) or not tolerance:
        return gdf
    if tolerance < PRECISION:
        raise ValueError(
            f"'tolerance' must be larger than {PRECISION} to avoid "
            "problems with floating point precision."
        )
    if duplicate_action not in ["fix", "error", "ignore"]:
        raise ValueError("duplicate_action must be 'fix', 'error' or 'ignore'")


def split_out_slivers(
    gdf: GeoDataFrame | GeoSeries, tolerance: float | int
) -> tuple[GeoDataFrame, GeoDataFrame] | tuple[GeoSeries, GeoSeries]:
    is_sliver = gdf.buffer(-tolerance / 2).is_empty
    slivers = gdf.loc[is_sliver]
    gdf = gdf.loc[~is_sliver]
    return gdf, slivers


def try_for_grid_size(
    func,
    grid_sizes: tuple[None, float | int],
    args: tuple | None = None,
    kwargs: dict | None = None,
) -> Any:
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}
    for i, grid_size in enumerate(grid_sizes):
        try:
            return func(*args, grid_size=grid_size, **kwargs)
        except GEOSException as e:
            if i == len(grid_sizes) - 1:
                raise e


def split_and_eliminate_by_longest(
    gdf: GeoDataFrame | list[GeoDataFrame],
    to_eliminate: GeoDataFrame,
    tolerance: int | float,
    grid_sizes: tuple[None | float | int] = (None,),
    logger=None,
    **kwargs,
) -> GeoDataFrame | tuple[GeoDataFrame]:
    if not len(to_eliminate):
        gdf = (gdf,) if isinstance(gdf, GeoDataFrame) else gdf
        return (*gdf, to_eliminate)

    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        as_gdf = pd.concat(gdf, ignore_index=True)
    else:
        as_gdf = gdf

    splitted = try_for_grid_size(
        split_by_neighbors,
        grid_sizes=grid_sizes,
        args=(to_eliminate, as_gdf, tolerance),
    ).pipe(sort_small_first)

    splitted = try_for_grid_size(
        update_geometries,
        grid_sizes=grid_sizes,
        args=(splitted,),
        kwargs=dict(geom_type="polygon"),
    )

    gdf = try_for_grid_size(
        eliminate_by_longest,
        grid_sizes=grid_sizes,
        args=(
            gdf,
            splitted,
        ),
        kwargs=kwargs,
    )

    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        as_gdf = pd.concat(gdf, ignore_index=True)
    else:
        as_gdf = gdf

    missing = try_for_grid_size(
        clean_overlay,
        grid_sizes=grid_sizes,
        args=(
            to_eliminate,
            as_gdf,
        ),
        kwargs=dict(
            how="difference",
            geom_type="polygon",
        ),
    ).pipe(dissexp_by_cluster)

    return try_for_grid_size(
        eliminate_by_longest, grid_sizes=grid_sizes, args=(gdf, missing), kwargs=kwargs
    )


def split_by_neighbors(df, split_by, tolerance, grid_size=None) -> GeoDataFrame:
    if not len(df):
        return df

    split_by = split_by.copy()
    split_by.geometry = shapely.simplify(split_by.geometry, tolerance)

    intersecting_lines = (
        clean_overlay(
            to_lines(split_by.explode(ignore_index=True)[lambda x: x.length > 0]),
            buff(df, tolerance),
            how="identity",
            grid_size=grid_size,
        )
        .pipe(get_line_segments)
        .reset_index(drop=True)
    )

    endpoints = intersecting_lines.boundary.explode(index_parts=False)

    extended_lines = GeoDataFrame(
        {
            "geometry": extend_lines(
                endpoints.loc[lambda x: ~x.index.duplicated(keep="first")].values,
                endpoints.loc[lambda x: ~x.index.duplicated(keep="last")].values,
                distance=tolerance * 3,
            )
        },
        crs=df.crs,
    )

    buffered = buff(extended_lines, tolerance, single_sided=True)

    return clean_overlay(
        df, buffered, how="identity", geom_type="polygon", grid_size=grid_size
    )


def extend_lines(arr1, arr2, distance) -> NDArray[LineString]:
    if len(arr1) != len(arr2):
        raise ValueError
    if not len(arr1):
        return arr1

    arr1, arr2 = arr2, arr1  # TODO fix

    coords1 = coordinate_array(arr1)
    coords2 = coordinate_array(arr2)

    dx = coords2[:, 0] - coords1[:, 0]
    dy = coords2[:, 1] - coords1[:, 1]
    len_xy = np.sqrt((dx**2.0) + (dy**2.0))
    x = coords1[:, 0] + (coords1[:, 0] - coords2[:, 0]) / len_xy * distance
    y = coords1[:, 1] + (coords1[:, 1] - coords2[:, 1]) / len_xy * distance

    new_points = np.array([None for _ in range(len(arr1))])
    new_points[~np.isnan(x)] = shapely.points(x[~np.isnan(x)], y[~np.isnan(x)])

    new_points[~np.isnan(x)] = make_lines_between_points(
        arr2[~np.isnan(x)], new_points[~np.isnan(x)]
    )
    return new_points


def make_lines_between_points(
    arr1: NDArray[Point], arr2: NDArray[Point]
) -> NDArray[LineString]:
    if arr1.shape != arr2.shape:
        raise ValueError(
            f"Arrays must have equal shape. Got {arr1.shape} and {arr2.shape}"
        )
    coords: pd.DataFrame = pd.concat(
        [
            pd.DataFrame(get_coordinates(arr1), columns=["x", "y"]),
            pd.DataFrame(get_coordinates(arr2), columns=["x", "y"]),
        ]
    ).sort_index()

    return linestrings(coords.values, indices=coords.index)


def get_line_segments(lines) -> GeoDataFrame:
    assert lines.index.is_unique
    if isinstance(lines, GeoDataFrame):
        multipoints = lines.assign(
            **{
                lines._geometry_column_name: force_2d(
                    extract_unique_points(lines.geometry.values)
                )
            }
        )
        return multipoints_to_line_segments(multipoints.geometry)

    multipoints = GeoSeries(extract_unique_points(lines.values), index=lines.index)

    return multipoints_to_line_segments(multipoints)


def multipoints_to_line_segments(multipoints: GeoSeries) -> GeoDataFrame:
    if not len(multipoints):
        return GeoDataFrame({"geometry": multipoints}, index=multipoints.index)

    try:
        crs = multipoints.crs
    except AttributeError:
        crs = None

    try:
        point_df = multipoints.explode(index_parts=False)
        if isinstance(point_df, GeoSeries):
            point_df = point_df.to_frame("geometry")
    except AttributeError:
        points, indices = get_parts(multipoints, return_index=True)
        if isinstance(multipoints.index, pd.MultiIndex):
            indices = pd.MultiIndex.from_arrays(indices, names=multipoints.index.names)

        point_df = pd.DataFrame({"geometry": GeometryArray(points)}, index=indices)

    point_df["next"] = point_df.groupby(level=0)["geometry"].shift(-1)

    first_points = point_df.loc[lambda x: ~x.index.duplicated(), "geometry"]
    is_last_point = point_df["next"].isna()

    point_df.loc[is_last_point, "next"] = first_points
    assert point_df["next"].notna().all()

    point_df["geometry"] = [
        LineString([x1, x2])
        for x1, x2 in zip(point_df["geometry"], point_df["next"], strict=False)
    ]
    return GeoDataFrame(point_df.drop(columns=["next"]), geometry="geometry", crs=crs)


def explore_geosexception(e: GEOSException, *gdfs, logger=None) -> None:
    from ..maps.maps import Explore
    from ..maps.maps import explore
    from .conversion import to_gdf

    pattern = r"(\d+\.\d+)\s+(\d+\.\d+)"

    matches = re.findall(pattern, str(e))
    coords_in_error_message = [(float(match[0]), float(match[1])) for match in matches]
    exception_point = to_gdf(coords_in_error_message, crs=gdfs[0].crs)
    if len(exception_point):
        exception_point["wkt"] = exception_point.to_wkt()
        if logger:
            logger.error(
                e, Explore(exception_point, *gdfs, mask=exception_point.buffer(100))
            )
        else:
            explore(exception_point, *gdfs, mask=exception_point.buffer(100))
    else:
        if logger:
            logger.error(e, Explore(*gdfs))
        else:
            explore(*gdfs)



import json
import numbers
import warnings
from collections.abc import Callable
from contextlib import contextmanager
from typing import Any

import joblib
import numpy as np
import pandas as pd
import rasterio
import shapely
from affine import Affine
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from rasterio import features
from rasterio.enums import MergeAlg
from shapely import Geometry
from shapely.geometry import shape

from ..geopandas_tools.conversion import to_bbox


def _get_res_from_bounds(
    obj: GeoDataFrame | GeoSeries | Geometry | tuple, shape: tuple[int, ...]
) -> tuple[int, int] | None:
    minx, miny, maxx, maxy = to_bbox(obj)
    try:
        height, width = shape[-2:]
    except IndexError:
        return None
    resx = (maxx - minx) / width
    resy = (maxy - miny) / height
    return resx, resy


def _get_transform_from_bounds(
    obj: GeoDataFrame | GeoSeries | Geometry | tuple, shape: tuple[int, ...]
) -> Affine:
    minx, miny, maxx, maxy = to_bbox(obj)
    if len(shape) in [2, 3]:
        height, width = shape[-2:]
    else:
        return None
        # raise ValueError(shape)
    return rasterio.transform.from_bounds(minx, miny, maxx, maxy, width, height)


def _res_as_tuple(res: int | float | tuple[int | float]) -> tuple[int | float]:
    return (res, res) if isinstance(res, numbers.Number) else res


def _get_shape_from_bounds(
    obj: GeoDataFrame | GeoSeries | Geometry | tuple,
    res: int,
    indexes: int | tuple[int],
) -> tuple[int, int]:
    resx, resy = _res_as_tuple(res)

    minx, miny, maxx, maxy = to_bbox(obj)

    # minx = math.floor(minx / res) * res
    # maxx = math.ceil(maxx / res) * res
    # miny = math.floor(miny / res) * res
    # maxy = math.ceil(maxy / res) * res

    # # Compute output array shape. We guarantee it will cover the output
    # # bounds completely
    # width = round((maxx - minx) // res)
    # height = round((maxy - miny) // res)

    # if not isinstance(indexes, int):
    #     return len(indexes), height, width
    # return height, width

    diffx = maxx - minx
    diffy = maxy - miny
    width = int(diffx / resx)
    height = int(diffy / resy)
    if not isinstance(indexes, int):
        return len(indexes), width, height
    return height, width


def _array_to_geojson(
    array: np.ndarray, transform: Affine, processes: int
) -> list[tuple]:
    if hasattr(array, "mask"):
        if isinstance(array.mask, np.ndarray):
            mask = array.mask == False
        else:
            mask = None
        array = array.data
    else:
        mask = None

    try:
        return _array_to_geojson_loop(array, transform, mask, processes)
    except ValueError:
        try:
            array = array.astype(np.float32)
            return _array_to_geojson_loop(array, transform, mask, processes)

        except Exception as err:
            raise err.__class__(f"{array.shape}: {err}") from err


def _array_to_geojson_loop(array, transform, mask, processes):
    if processes == 1:
        return [
            (value, shape(geom))
            for geom, value in features.shapes(array, transform=transform, mask=mask)
        ]
    else:
        with joblib.Parallel(n_jobs=processes, backend="threading") as parallel:
            return parallel(
                joblib.delayed(_value_geom_pair)(value, geom)
                for geom, value in features.shapes(
                    array, transform=transform, mask=mask
                )
            )


def _value_geom_pair(value, geom):
    return (value, shape(geom))


def _gdf_to_arr(
    gdf: GeoDataFrame,
    res: int | float | None = None,
    out_shape: tuple[int, int] | None = None,
    bounds: tuple[float] | None = None,
    fill: int = 0,
    all_touched: bool = False,
    merge_alg: Callable = MergeAlg.replace,
    default_value: int = 1,
    dtype: Any | None = None,
) -> np.ndarray:
    """Construct Raster from a GeoDataFrame or GeoSeries.

    The GeoDataFrame should have

    Args:
        gdf: The GeoDataFrame to rasterize.
        res: Resolution of the raster in units of the GeoDataFrame's coordinate reference system.
        bounds: Optional bounds to box 'gdf' into (so both clip and extend to).
        fill: Fill value for areas outside of input geometries (default is 0).
        all_touched: Whether to consider all pixels touched by geometries,
            not just those whose center is within the polygon (default is False).
        merge_alg: Merge algorithm to use when combining geometries
            (default is 'MergeAlg.replace').
        default_value: Default value to use for the rasterized pixels
            (default is 1).
        dtype: Data type of the output array. If None, it will be
            determined automatically.
        out_shape: Optional 2 dimensional shape of the resulting array.

    Returns:
        A Raster instance based on the specified GeoDataFrame and parameters.

    Raises:
        TypeError: If 'transform' is provided in kwargs, as this is
        computed based on the GeoDataFrame bounds and resolution.
    """
    if res is not None and out_shape is not None:
        raise TypeError("Cannot specify both 'res' and 'out_shape'")
    if res is None and out_shape is None:
        raise TypeError("Must specify either 'res' or 'out_shape'")

    if isinstance(gdf, GeoSeries):
        gdf = gdf.to_frame("geometry")
    elif not isinstance(gdf, GeoDataFrame):
        raise TypeError(type(gdf))

    if bounds is not None:
        gdf = gdf.clip(bounds)
        bounds_gdf = GeoDataFrame({"geometry": [shapely.box(*bounds)]}, crs=gdf.crs)

    if len(gdf.columns) > 2:
        raise ValueError(
            "gdf should have only a geometry column and one numeric column to "
            "use as array values. "
            "Alternatively only a geometry column and a numeric index."
        )
    elif len(gdf.columns) == 1:
        values = np.full(len(gdf), default_value)
    else:
        col: str = next(iter(gdf.columns.difference({gdf.geometry.name})))
        values = gdf[col].values

    if bounds is not None:
        gdf = pd.concat([bounds_gdf, gdf])
        values = np.concatenate([np.array([fill]), values])

    if out_shape is None:
        assert res is not None
        out_shape = _get_shape_from_bounds(gdf.total_bounds, res=res, indexes=1)

    if not len(gdf):
        return np.full(out_shape, fill)

    transform = _get_transform_from_bounds(gdf.total_bounds, out_shape)

    return features.rasterize(
        _gdf_to_geojson_with_col(gdf, values),
        out_shape=out_shape,
        transform=transform,
        fill=fill,
        all_touched=all_touched,
        merge_alg=merge_alg,
        default_value=default_value,
        dtype=dtype,
    )


def _gdf_to_geojson_with_col(gdf: GeoDataFrame, values: np.ndarray) -> list[dict]:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        return [
            (feature["geometry"], val)
            for val, feature in zip(
                values, json.loads(gdf.to_json())["features"], strict=False
            )
        ]


@contextmanager
def memfile_from_array(array: np.ndarray, **profile) -> rasterio.MemoryFile:
    """Yield a memory file from a numpy array."""
    with rasterio.MemoryFile() as memfile:
        with memfile.open(**profile) as dataset:
            dataset.write(array, indexes=profile["indexes"])
        with memfile.open() as dataset:
            yield dataset


def get_index_mapper(df: pd.DataFrame) -> tuple[dict[int, int], str]:
    """Get a dict of index mapping and the name of the index."""
    idx_mapper = dict(enumerate(df.index))
    idx_name = df.index.name
    return idx_mapper, idx_name



from typing import Any

try:
    from gcsfs import GCSFileSystem

    file_system = GCSFileSystem
except ImportError:
    import datetime
    import glob
    import io
    import os
    import pathlib
    from concurrent.futures import ThreadPoolExecutor
    from typing import Any

    class LocalFileSystem:
        """Mimicks GCS's FileSystem but using standard library (os, glob)."""

        @staticmethod
        def glob(
            path: str,
            detail: bool = False,
            recursive: bool = True,
            include_hidden: bool = True,
            **kwargs,
        ) -> list[dict] | list[str]:
            """Like GCSFileSystem.glob."""
            relevant_paths = glob.iglob(
                path, recursive=recursive, include_hidden=include_hidden, **kwargs
            )

            if not detail:
                return list(relevant_paths)
            with ThreadPoolExecutor() as executor:
                return list(executor.map(get_file_info, relevant_paths))

        @classmethod
        def ls(
            cls, path: str, detail: bool = False, **kwargs
        ) -> list[dict] | list[str]:
            """Like GCSFileSystem.ls."""
            return cls().glob(
                str(pathlib.Path(path) / "**"), detail=detail, recursive=False, **kwargs
            )

        @staticmethod
        def info(path) -> dict[str, Any]:
            """Like GCSFileSystem.info."""
            return get_file_info(path)

        @staticmethod
        def open(path: str, *args, **kwargs) -> io.TextIOWrapper:
            """Built in open."""
            return open(path, *args, **kwargs)

        @staticmethod
        def exists(path: str) -> bool:
            """os.path.exists."""
            return os.path.exists(path)

    def get_file_info(path) -> dict[str, str | float]:
        return {
            "updated": datetime.datetime.fromtimestamp(os.path.getmtime(path)),
            "size": os.path.getsize(path),
            "name": path,
            "type": "directory" if os.path.isdir(path) else "file",
        }

    file_system = LocalFileSystem

from .geopandas_tools.runners import OverlayRunner
from .geopandas_tools.runners import RTreeQueryRunner
from .geopandas_tools.runners import UnionRunner


def _get_instance(data: dict, key: str, **kwargs) -> Any:
    """Get the dict value and call it if callable."""
    x = data[key]
    if callable(x):
        return x(**kwargs)
    return x


config = {
    "n_jobs": 1,
    "file_system": file_system,
    "rtree_runner": RTreeQueryRunner,
    "overlay_runner": OverlayRunner,
    "union_runner": UnionRunner,
}



"""Create nodes (points) from a GeoDataFrame of lines.

The functions are used inside inside other functions, and aren't needed
to use explicitly.
"""

import geopandas as gpd
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from shapely.geometry import Point

from ..geopandas_tools.general import _push_geom_col
from ..geopandas_tools.general import make_edge_coords_cols
from ..geopandas_tools.general import make_edge_wkt_cols
from ..geopandas_tools.geometry_types import make_all_singlepart


def make_node_ids(
    gdf: GeoDataFrame,
    wkt: bool = True,
) -> tuple[GeoDataFrame, GeoDataFrame]:
    """Gives the lines unique node ids and returns lines (edges) and nodes.

    Takes the first and last point of each line and creates a GeoDataFrame of
    nodes (points) with a column 'node_id'. The node ids are then assigned to the
    input GeoDataFrame of lines as the columns 'source' and 'target'.

    Args:
        gdf: GeoDataFrame with line geometries
        wkt: If True (default), the resulting nodes will include the column 'wkt',
            containing the well-known text representation of the geometry. If False, it
            will include the column 'coords', a tuple with x and y geometries.

    Returns:
        A tuple of two GeoDataFrames, one with the lines and one with the nodes.

    Note:
        The lines must be singlepart linestrings.
    """
    gdf = make_all_singlepart(gdf, ignore_index=True)

    if wkt:
        gdf = make_edge_wkt_cols(gdf)
        geomcol1, geomcol2, geomcol_final = "source_wkt", "target_wkt", "wkt"
    else:
        gdf = make_edge_coords_cols(gdf)
        geomcol1, geomcol2, geomcol_final = "source_coords", "target_coords", "coords"

    # remove identical lines in opposite directions
    gdf["meters_"] = gdf.length.astype(str)

    sources = gdf[[geomcol1, geomcol2, "meters_"]].rename(
        columns={geomcol1: geomcol_final, geomcol2: "temp"}
    )
    targets = gdf[[geomcol1, geomcol2, "meters_"]].rename(
        columns={geomcol2: geomcol_final, geomcol1: "temp"}
    )

    nodes = (
        pd.concat([sources, targets], axis=0, ignore_index=True)
        .drop_duplicates([geomcol_final, "temp", "meters_"])
        .drop(["meters_", "temp"], axis=1)
    )

    gdf = gdf.drop("meters_", axis=1)

    nodes["n"] = nodes.assign(n=1).groupby(geomcol_final)["n"].transform("sum")

    nodes = nodes.drop_duplicates(subset=[geomcol_final]).reset_index(drop=True)

    nodes["node_id"] = nodes.index
    nodes["node_id"] = nodes["node_id"].astype(str)

    id_dict = {
        geom: node_id
        for geom, node_id in zip(nodes[geomcol_final], nodes["node_id"], strict=True)
    }
    gdf["source"] = gdf[geomcol1].map(id_dict)
    gdf["target"] = gdf[geomcol2].map(id_dict)

    n_dict = {geom: n for geom, n in zip(nodes[geomcol_final], nodes["n"], strict=True)}
    gdf["n_source"] = gdf[geomcol1].map(n_dict)
    gdf["n_target"] = gdf[geomcol2].map(n_dict)

    if wkt:
        nodes["geometry"] = gpd.GeoSeries.from_wkt(nodes[geomcol_final], crs=gdf.crs)
    else:
        nodes["geometry"] = GeoSeries(
            [Point(geom) for geom in nodes[geomcol_final]], crs=gdf.crs
        )
    nodes = gpd.GeoDataFrame(nodes, geometry="geometry", crs=gdf.crs)
    nodes = nodes.reset_index(drop=True)

    gdf = _push_geom_col(gdf)

    return gdf, nodes



"""Functions for reading and writing GeoDataFrames in Statistics Norway's GCS Dapla."""

from __future__ import annotations

import functools
import glob
import json
import multiprocessing
import os
import shutil
import uuid
from collections.abc import Callable
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from io import BytesIO
from pathlib import Path
from typing import Any

import geopandas as gpd
import joblib
import pandas as pd
import pyarrow
import pyarrow.dataset
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import shapely
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from geopandas.io.arrow import _arrow_to_geopandas
from geopandas.io.arrow import _geopandas_to_arrow
from pandas import DataFrame
from pyarrow import ArrowInvalid

from ..conf import config
from ..geopandas_tools.conversion import to_shapely
from ..geopandas_tools.general import get_common_crs
from ..geopandas_tools.sfilter import sfilter
from ..helpers import _get_file_system

try:
    from gcsfs import GCSFileSystem
except ImportError:

    class GCSFileSystem:
        """Placeholder."""


PANDAS_FALLBACK_INFO = " Set pandas_fallback=True to ignore this error."
NULL_VALUE = "__HIVE_DEFAULT_PARTITION__"


def read_geopandas(
    gcs_path: str | Path | list[str | Path] | tuple[str | Path] | GeoSeries,
    pandas_fallback: bool = False,
    file_system: GCSFileSystem | None = None,
    mask: GeoSeries | GeoDataFrame | shapely.Geometry | tuple | None = None,
    use_threads: bool = True,
    filters: pyarrow.dataset.Expression | None = None,
    **kwargs,
) -> GeoDataFrame | DataFrame:
    """Reads geoparquet or other geodata from one or more files on GCS.

    If the file has 0 rows, the contents will be returned as a pandas.DataFrame,
    since geopandas does not read and write empty tables.

    Note:
        Does not currently read shapefiles or filegeodatabases.

    Args:
        gcs_path: path to one or more files on Google Cloud Storage.
            Multiple paths are read with threading.
        pandas_fallback: If False (default), an exception is raised if the file can
            not be read with geopandas and the number of rows is more than 0. If True,
            the file will be read with pandas if geopandas fails.
        file_system: Optional file system.
        mask: If gcs_path is a partitioned parquet file or an interable of paths.
            Only files with a bbox intersecting mask will be read.
            Note that the data is not filtered on a row level. You should either
            use clip or sfilter to filter the data after reading.
        use_threads: Defaults to True.
        filters: To filter out data. Either a pyarrow.dataset.Expression, or a list in the
            structure [[(column, op, val), …],…] where op is [==, =, >, >=, <, <=, !=, in, not in].
            More details here: https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html
        **kwargs: Additional keyword arguments passed to geopandas' read_parquet
            or read_file, depending on the file type.

    Returns:
         A GeoDataFrame if it has rows. If zero rows, a pandas DataFrame is returned.
    """
    file_system = _get_file_system(file_system, kwargs)

    if isinstance(gcs_path, (Path | os.PathLike)):
        gcs_path = str(gcs_path)

    if isinstance(gcs_path, str):
        gcs_path = _maybe_strip_prefix(gcs_path, file_system)
    else:
        gcs_path = [_maybe_strip_prefix(str(path), file_system) for path in gcs_path]

        return _read_geopandas_from_iterable(
            gcs_path,
            mask=mask,
            file_system=file_system,
            use_threads=use_threads,
            pandas_fallback=pandas_fallback,
            filters=filters,
            **kwargs,
        )

    single_eq_filter = (
        isinstance(filters, Iterable)
        and len(filters) == 1
        and ("=" in next(iter(filters)) or "==" in next(iter(filters)))
    )
    # try to read files in subfolder path / "column=value"
    # because glob is slow without GCSFileSystem from the root partition
    if single_eq_filter:
        try:
            expression: list[str] = "".join(
                [str(x) for x in next(iter(filters))]
            ).replace("==", "=")
            glob_func = _get_glob_func(file_system)
            suffix: str = Path(gcs_path).suffix
            paths = glob_func(str(Path(gcs_path) / expression / f"*{suffix}"))
            if paths:
                return _read_geopandas_from_iterable(
                    paths,
                    mask=mask,
                    file_system=file_system,
                    use_threads=use_threads,
                    pandas_fallback=pandas_fallback,
                    filters=filters,
                    **kwargs,
                )
        except FileNotFoundError:
            pass

    child_paths = get_child_paths(gcs_path, file_system)
    if child_paths:
        return gpd.GeoDataFrame(
            _read_partitioned_parquet(
                gcs_path,
                file_system=file_system,
                mask=mask,
                filters=filters,
                child_paths=child_paths,
                use_threads=use_threads,
                **kwargs,
            )
        )

    if gcs_path.endswith(".parquet"):
        file_format: str = "parquet"
        read_func = gpd.read_parquet
    else:
        file_format: str = Path(gcs_path).suffix.lstrip(".")
        read_func = gpd.read_file

    with file_system.open(gcs_path, mode="rb") as file:
        return _read_geopandas_single_path(
            file,
            read_func=read_func,
            file_format=file_format,
            filters=filters,
            **kwargs,
        )


def _read_geopandas_from_iterable(
    paths, mask, file_system, use_threads, pandas_fallback, **kwargs
):
    cols = {}
    if mask is None and isinstance(paths, GeoSeries):
        # bounds GeoSeries indexed with file paths
        paths = list(paths.index)
    elif mask is None:
        paths = list(paths)
    elif isinstance(paths, GeoSeries):
        bounds_series = sfilter(paths, mask)
        if not len(bounds_series):
            # return GeoDataFrame with correct columns
            if isinstance(kwargs.get("columns"), Iterable):
                cols = {col: [] for col in kwargs["columns"]}
            else:
                cols = {}
                for path in bounds_series.index:
                    try:
                        cols |= {col: [] for col in _get_columns(path, file_system)}
                    except ArrowInvalid as e:
                        if file_system.isfile(path):
                            raise ArrowInvalid(e, path) from e
            first_path = next(iter(paths.index))
            _, crs = _get_bounds_parquet(first_path, file_system)
            return GeoDataFrame(cols | {"geometry": []}, crs=crs)
        paths = list(bounds_series.index)

    results: list[pyarrow.Table] = _read_pyarrow_with_treads(
        paths,
        file_system=file_system,
        mask=mask,
        use_threads=use_threads,
        **kwargs,
    )
    if results:
        try:
            return _concat_pyarrow_to_geopandas(results, paths, file_system)
        except Exception as e:
            if not pandas_fallback:
                print(e)
                raise e
    else:
        first_path = next(iter(paths))
        _, crs = _get_bounds_parquet(first_path, file_system)
        df = GeoDataFrame(cols | {"geometry": []}, crs=crs)

    return df


def _read_pyarrow_with_treads(
    paths: list[str | Path | os.PathLike],
    file_system,
    use_threads,
    mask,
    filters,
    **kwargs,
) -> list[pyarrow.Table]:
    read_partial = functools.partial(
        _read_pyarrow, filters=filters, mask=mask, file_system=file_system, **kwargs
    )
    if not use_threads:
        return [x for x in map(read_partial, paths) if x is not None]
    with ThreadPoolExecutor() as executor:
        return [x for x in executor.map(read_partial, paths) if x is not None]


def intersects(file, mask, file_system) -> bool:
    bbox, _ = _get_bounds_parquet_from_open_file(file, file_system)
    return shapely.box(*bbox).intersects(to_shapely(mask))


def _read_pyarrow(path: str, file_system, mask=None, **kwargs) -> pyarrow.Table | None:
    try:
        with file_system.open(path, "rb") as file:
            if mask is not None and not intersects(file, mask, file_system):
                return

            # 'get' instead of 'pop' because dict is mutable
            schema = kwargs.get("schema", pq.read_schema(file))
            new_kwargs = {
                key: value for key, value in kwargs.items() if key != "schema"
            }

            return pq.read_table(file, schema=schema, **new_kwargs)
    except ArrowInvalid as e:
        glob_func = _get_glob_func(file_system)
        if not len(
            {
                x
                for x in glob_func(str(Path(path) / "**"))
                if not paths_are_equal(path, x)
            }
        ):
            raise e
        # allow not being able to read empty directories that are hard to delete in gcs


def _get_bounds_parquet(
    path: str | Path, file_system: GCSFileSystem, pandas_fallback: bool = False
) -> tuple[list[float], dict] | tuple[None, None]:
    with file_system.open(path, "rb") as file:
        return _get_bounds_parquet_from_open_file(file, file_system)


def _get_bounds_parquet_from_open_file(
    file, file_system
) -> tuple[list[float], dict] | tuple[None, None]:
    geo_metadata = _get_geo_metadata_primary_column(file, file_system)

    if not geo_metadata:
        return None, None
    return geo_metadata["bbox"], geo_metadata["crs"]


def _get_geo_metadata(file, file_system) -> dict:
    try:
        meta = pq.read_schema(file).metadata
    except FileNotFoundError:
        try:
            with file_system.open(file, "rb") as f:
                meta = pq.read_schema(f).metadata
        except Exception as e:
            raise e.__class__(f"{file}: {e}") from e
    except Exception as e:
        raise e.__class__(f"{file}: {e}") from e

    return json.loads(meta[b"geo"])


def _get_geo_metadata_primary_column(file, file_system) -> dict:
    geo_metadata = _get_geo_metadata(file, file_system)
    try:
        primary_column = geo_metadata["primary_column"]
    except KeyError as e:
        raise KeyError(e, geo_metadata) from e
    try:
        return geo_metadata["columns"][primary_column]
    except KeyError as e:
        try:
            num_rows = pq.read_metadata(file).num_rows
        except ArrowInvalid as e:
            if not file_system.isfile(file):
                return {}
            raise ArrowInvalid(e, file) from e
        # allow for 0 lengthed tables not to have geo metadata
        if not num_rows:
            return {}
    return {}


def _get_columns(path: str | Path, file_system: GCSFileSystem) -> pd.Index:
    with file_system.open(path, "rb") as f:
        schema = pq.read_schema(f)
        index_cols = _get_index_cols(schema)
        return pd.Index(schema.names).difference(index_cols)


def _get_index_cols(schema: pyarrow.Schema) -> list[str]:
    cols = json.loads(schema.metadata[b"pandas"])["index_columns"]
    return [x for x in cols if not isinstance(x, dict)]


def get_bounds_series(
    paths: list[str | Path] | tuple[str | Path],
    file_system: GCSFileSystem | None = None,
    use_threads: bool = True,
    pandas_fallback: bool = False,
) -> GeoSeries:
    """Get a GeoSeries with file paths as indexes and the file's bounds as values.

    The returned GeoSeries can be used as the first argument of 'read_geopandas'
    along with the 'mask' keyword.

    Args:
        paths: Iterable of file paths in gcs.
        file_system: Optional instance of GCSFileSystem.
            If None, an instance is created within the function.
            Note that this is slower in long loops.
        use_threads: Default True.
        pandas_fallback: If False (default), an exception is raised if the file has
            no geo metadata. If True, the geometry value is set to None for this file.

    Returns:
        A geopandas.GeoSeries with file paths as indexes and bounds as values.

    Examples:
    ---------
    >>> import sgis as sg
    >>> import dapla as dp
    >>> all_paths =  GCSFileSystem().ls("...")

    Get the bounds of all your file paths, indexed by path.

    >>> bounds_series = sg.get_bounds_series(all_paths, file_system)
    >>> bounds_series
    .../0301.parquet    POLYGON ((273514.334 6638380.233, 273514.334 6...
    .../1101.parquet    POLYGON ((6464.463 6503547.192, 6464.463 65299...
    .../1103.parquet    POLYGON ((-6282.301 6564097.347, -6282.301 660...
    .../1106.parquet    POLYGON ((-46359.891 6622984.385, -46359.891 6...
    .../1108.parquet    POLYGON ((30490.798 6551661.467, 30490.798 658...
                                                                                                            ...
    .../5628.parquet    POLYGON ((1019391.867 7809550.777, 1019391.867...
    .../5630.parquet    POLYGON ((1017907.145 7893398.317, 1017907.145...
    .../5632.parquet    POLYGON ((1075687.587 7887714.263, 1075687.587...
    .../5634.parquet    POLYGON ((1103447.451 7874551.663, 1103447.451...
    .../5636.parquet    POLYGON ((1024129.618 7838961.91, 1024129.618 ...
    Length: 357, dtype: geometry

    Make a grid around the total bounds of the files,
    and read geometries intersecting with the mask in a loop.

    >>> grid = sg.make_grid(bounds_series, 10_000)
    >>> for mask in grid.geometry:
    ...     df = sg.read_geopandas(
    ...         bounds_series,
    ...         mask=mask,
    ...         file_system=file_system,
    ...     )

    """
    file_system = _get_file_system(file_system, {})

    if isinstance(paths, (str | Path)):
        paths = [paths]

    threads = (
        min(len(paths), int(multiprocessing.cpu_count())) or 1 if use_threads else 1
    )

    with joblib.Parallel(n_jobs=threads, backend="threading") as parallel:
        bounds: list[tuple[list[float], dict]] = parallel(
            joblib.delayed(_get_bounds_parquet)(
                path, file_system=file_system, pandas_fallback=pandas_fallback
            )
            for path in paths
        )
    crss = {json.dumps(x[1]) for x in bounds}
    crs = get_common_crs(
        [
            crs
            for crs in crss
            if not any(str(crs).lower() == txt for txt in ["none", "null"])
        ]
    )
    return GeoSeries(
        [shapely.box(*bbox[0]) if bbox[0] is not None else None for bbox in bounds],
        index=paths,
        crs=crs,
    )


def write_geopandas(
    df: GeoDataFrame,
    gcs_path: str | Path,
    overwrite: bool = True,
    pandas_fallback: bool = False,
    file_system: GCSFileSystem | None = None,
    partition_cols=None,
    existing_data_behavior: str = "error",
    **kwargs,
) -> None:
    """Writes a GeoDataFrame to the speficied format.

    Note:
        Does not currently write to shapelfile or filegeodatabase.

    Args:
        df: The GeoDataFrame to write.
        gcs_path: The path to the file you want to write to.
        overwrite: Whether to overwrite the file if it exists. Defaults to True.
        pandas_fallback: If False (default), an exception is raised if the file can
            not be written with geopandas and the number of rows is more than 0. If True,
            the file will be written without geo-metadata if >0 rows.
        file_system: Optional file sustem.
        partition_cols: Column(s) to partition by. Only for parquet files.
        existing_data_behavior : 'error' | 'overwrite_or_ignore' | 'delete_matching'.
            Defaults to 'error'. More info: https://arrow.apache.org/docs/python/generated/pyarrow.dataset.write_dataset.html
        **kwargs: Additional keyword arguments passed to parquet.write_table
            (for parquet) or geopandas' to_file method (if not parquet).
    """
    if not isinstance(gcs_path, str):
        try:
            gcs_path = str(gcs_path)
        except TypeError as e:
            raise TypeError(f"Unexpected type {type(gcs_path)}.") from e

    file_system = _get_file_system(file_system, kwargs)

    if not overwrite and file_system.exists(gcs_path):
        raise ValueError("File already exists.")

    if not isinstance(df, GeoDataFrame):
        raise ValueError(
            f"DataFrame must be GeoDataFrame. Got {type(df)} for {gcs_path}."
        )

    if not len(df) and get_child_paths(gcs_path, file_system):
        # no need to write empty df for partitioned parquet - if root dir exists
        return
    elif not len(df):
        if pandas_fallback:
            df = pd.DataFrame(df)
            df.geometry = df.geometry.astype(str)
            df.geometry = None
        try:
            file_format: str = Path(gcs_path).suffix.lstrip(".")
            write_method: Callable = getattr(df, f"to_{file_format}")
            if file_format == "parquet":
                kwargs["engine"] = "pyarrow"
            with file_system.open(gcs_path, "wb") as file:
                write_method(file, **kwargs)

        except Exception as e:
            more_txt = PANDAS_FALLBACK_INFO if not pandas_fallback else ""
            raise e.__class__(
                f"{e.__class__.__name__}: {e} for {df}. " + more_txt
            ) from e
        return

    if ".parquet" in gcs_path or "prqt" in gcs_path:
        if partition_cols is not None:
            return _write_partitioned_geoparquet(
                df,
                gcs_path,
                partition_cols,
                file_system,
                existing_data_behavior=existing_data_behavior,
                write_func=_to_geopandas,
                **kwargs,
            )
        with file_system.open(gcs_path, mode="wb") as file:
            _to_geopandas(df, file, **kwargs)
        return

    layer = kwargs.pop("layer", None)
    if ".gpkg" in gcs_path:
        driver = "GPKG"
        layer = Path(gcs_path).stem
    elif ".geojson" in gcs_path:
        driver = "GeoJSON"
    elif ".gml" in gcs_path:
        driver = "GML"
    elif ".shp" in gcs_path:
        driver = "ESRI Shapefile"
    else:
        driver = None

    with BytesIO() as buffer:
        df.to_file(buffer, driver=driver)
        buffer.seek(0)  # Rewind the buffer to the beginning

        # Upload buffer content to the desired storage
        with file_system.open(gcs_path, "wb") as file:
            file.write(buffer.read())


def _to_geopandas(df, path, **kwargs) -> None:
    table = _geopandas_to_arrow(
        df,
        index=df.index,
        schema_version=None,
    )

    if "schema" in kwargs:
        schema = kwargs.pop("schema")

        # make sure to get the actual metadata
        schema = pyarrow.schema(
            [(schema.field(col).name, schema.field(col).type) for col in schema.names],
            metadata=table.schema.metadata,
        )
        table = table.select(schema.names).cast(schema)

    pq.write_table(table, path, compression="snappy", **kwargs)


def _pyarrow_schema_from_geopandas(df: GeoDataFrame) -> pyarrow.Schema:
    geom_name = df.geometry.name
    pandas_columns = [col for col in df if col != geom_name]
    schema = pyarrow.Schema.from_pandas(df[pandas_columns], preserve_index=True)
    index_columns = _get_index_cols(schema)
    return pyarrow.schema(
        [
            (
                (schema.field(col).name, schema.field(col).type)
                if col != geom_name
                else (geom_name, pyarrow.binary())
            )
            for col in [*df.columns, *index_columns]
        ]
    )


def _remove_file(path, file_system) -> None:
    try:
        file_system.rm_file(str(path))
    except (AttributeError, TypeError, PermissionError) as e:
        print(path, type(e), e)
        try:
            shutil.rmtree(path)
        except NotADirectoryError:
            try:
                os.remove(path)
            except PermissionError:
                pass


def _write_partitioned_geoparquet(
    df,
    path,
    partition_cols,
    file_system=None,
    write_func: Callable = _to_geopandas,
    existing_data_behavior: str = "error",
    basename_template: str | None = None,
    **kwargs,
):
    file_system = _get_file_system(file_system, kwargs)

    if isinstance(partition_cols, str):
        partition_cols = [partition_cols]

    for col in partition_cols:
        if df[col].isna().all() and not kwargs.get("schema"):
            raise ValueError("Must specify 'schema' when all rows are NA.")

    glob_func = _get_glob_func(file_system)

    if file_system.exists(path) and file_system.isfile(path):
        _remove_file(path, file_system)

    if kwargs.get("schema"):
        schema = kwargs.pop("schema")
    elif isinstance(df, GeoDataFrame):
        schema = _pyarrow_schema_from_geopandas(df)
    else:
        schema = pyarrow.Schema.from_pandas(df, preserve_index=True)

    def as_partition_part(col: str, value: Any) -> str:
        value = value if not pd.isna(value) else NULL_VALUE
        return f"{col}={value}"

    paths: list[Path] = []
    dfs: list[DataFrame] = []
    for group, rows in df.groupby(partition_cols, dropna=False):
        partition_parts = "/".join(
            as_partition_part(col, value)
            for col, value in zip(partition_cols, group, strict=True)
        )
        paths.append(Path(path) / partition_parts)
        dfs.append(rows)

    def threaded_write(rows: DataFrame, path: str) -> None:
        if basename_template is None:
            this_basename = (uuid.uuid4().hex + "-{i}.parquet").replace("-{i}", "0")
        else:
            this_basename = basename_template
        for i, sibling_path in enumerate(sorted(glob_func(str(Path(path) / "**")))):
            if paths_are_equal(sibling_path, path):
                continue
            if existing_data_behavior == "delete_matching":
                _remove_file(sibling_path, file_system)
            elif existing_data_behavior == "error":
                raise pyarrow.ArrowInvalid(
                    f"Could not write to {path} as the directory is not empty and existing_data_behavior is to error"
                )
            else:
                this_basename = basename_template.replace("-{i}", str(i + 1))

        out_path = str(Path(path) / this_basename)
        try:
            with file_system.open(out_path, mode="wb") as file:
                write_func(rows, file, schema=schema, **kwargs)
        except FileNotFoundError:
            file_system.makedirs(str(path), exist_ok=True)
            with file_system.open(out_path, mode="wb") as file:
                write_func(rows, file, schema=schema, **kwargs)

    with ThreadPoolExecutor() as executor:
        executor.map(threaded_write, dfs, paths)


def _get_glob_func(file_system) -> functools.partial:
    try:
        return functools.partial(file_system.glob)
    except AttributeError:
        return functools.partial(glob.glob, recursive=True)


def _filters_to_expression(filters) -> list[ds.Expression]:
    if filters is None:
        return None
    elif isinstance(filters, pyarrow.dataset.Expression):
        return filters

    for filt in filters:
        if "in" in filt and isinstance(filt[-1], str):
            raise ValueError(
                "Using strings with 'in' is ambigous. Use a list of strings."
            )
    try:
        return pq.core.filters_to_expression(filters)
    except ValueError as e:
        raise ValueError(f"{e}: {filters}") from e


def expression_match_path(expression: ds.Expression, path: str) -> bool:
    """Check if a file path match a pyarrow Expression.

    Examples:
    ---------
    >>> import pyarrow.compute as pc
    >>> path = 'data/file.parquet/x=1/y=10/name0.parquet'
    >>> expression = (pc.Field("x") == 1) & (pc.Field("y") == 10)
    >>> expression_match_path(path, expression)
    True
    >>> expression = (pc.Field("x") == 1) & (pc.Field("y") == 5)
    >>> expression_match_path(path, expression)
    False
    >>> expression = (pc.Field("x") == 1) & (pc.Field("z") == 10)
    >>> expression_match_path(path, expression)
    False
    """
    # keep only the parts in between the two .parquet parts
    path = str(path).split(".parquet")[1]
    if NULL_VALUE in path:
        return True
    # build a one lengthed pyarrow.Table of the partitioning in the file path
    values = []
    names = []
    for part in Path(path).parts:
        if part.count("=") != 1:
            continue
        name, value = part.split("=")
        values.append([value])
        names.append(name)
    table = pyarrow.Table.from_arrays(values, names=names)
    try:
        table = table.filter(expression)
    except pyarrow.ArrowInvalid as e:
        if "No match for FieldRef" in str(e):
            # if a non-partition col is used in 'filters',
            # we cannot determine if the expression match without reading the file
            return True
        raise e
    return bool(len(table))


def _read_geopandas_single_path(
    file,
    read_func: Callable,
    file_format: str,
    **kwargs,
):
    try:
        return read_func(file, **kwargs)
    except ValueError as e:
        if "Missing geo metadata" not in str(e) and "geometry" not in str(e):
            raise e.__class__(f"{e.__class__.__name__}: {e} for {file}. ") from e
        df = getattr(pd, f"read_{file_format}")(file, **kwargs)
        if not len(df):
            return GeoDataFrame(df)
        raise e.__class__(f"{e.__class__.__name__}: {e} for {df}. ") from e
    except Exception as e:
        raise e.__class__(f"{e.__class__.__name__}: {e} for {file}.") from e


def _read_pandas(gcs_path: str, use_threads: bool = True, **kwargs):
    file_system = _get_file_system(None, kwargs)

    if not isinstance(gcs_path, (str | Path | os.PathLike)):
        results: list[pyarrow.Table] = _read_pyarrow_with_treads(
            gcs_path,
            file_system=file_system,
            mask=None,
            use_threads=use_threads,
            **kwargs,
        )
        results = pyarrow.concat_tables(results, promote_options="permissive")
        return results.to_pandas()

    child_paths = get_child_paths(gcs_path, file_system)
    if child_paths:
        return _read_partitioned_parquet(
            gcs_path,
            file_system=file_system,
            mask=None,
            child_paths=child_paths,
            use_threads=use_threads,
            to_geopandas=False,
            **kwargs,
        )

    with file_system.open(gcs_path, "rb") as file:
        return pd.read_parquet(file, **kwargs)


def _read_partitioned_parquet(
    path: str,
    filters=None,
    file_system=None,
    mask=None,
    child_paths: list[str] | None = None,
    use_threads: bool = True,
    to_geopandas: bool = True,
    **kwargs,
):
    file_system = _get_file_system(file_system, kwargs)
    glob_func = _get_glob_func(file_system)

    if child_paths is None:
        child_paths = list(glob_func(str(Path(path) / "**/*.parquet")))

    filters = _filters_to_expression(filters)

    results: list[pyarrow.Table] = _read_pyarrow_with_treads(
        (
            path
            for path in child_paths
            if filters is None or expression_match_path(filters, path)
        ),
        file_system=file_system,
        mask=mask,
        filters=filters,
        use_threads=use_threads,
        **kwargs,
    )

    if results and to_geopandas:
        return _concat_pyarrow_to_geopandas(results, child_paths, file_system)
    elif results:
        return pyarrow.concat_tables(results, promote_options="permissive").to_pandas()

    # add columns to empty DataFrame
    first_path = next(iter(child_paths + [path]))
    _, crs = _get_bounds_parquet(first_path, file_system)
    df = GeoDataFrame(columns=_get_columns(first_path, file_system), crs=crs)
    if kwargs.get("columns"):
        return df[list(kwargs["columns"])]
    return df


def _concat_pyarrow_to_geopandas(
    results: list[pyarrow.Table], paths: list[str], file_system: Any
):
    results = pyarrow.concat_tables(
        results,
        promote_options="permissive",
    )
    geo_metadata = _get_geo_metadata(next(iter(paths)), file_system)
    return _arrow_to_geopandas(results, geo_metadata)


def paths_are_equal(path1: Path | str, path2: Path | str) -> bool:
    return Path(path1).parts == Path(path2).parts


def get_child_paths(path, file_system) -> list[str]:
    glob_func = _get_glob_func(file_system)
    return [
        x
        for x in glob_func(str(Path(path) / "**/*.parquet"))
        if not paths_are_equal(x, path)
    ]


def check_files(
    folder: str,
    contains: str | None = None,
    within_minutes: int | None = None,
) -> pd.DataFrame:
    """Returns DataFrame of files in the folder and subfolders with times and sizes.

    Args:
        folder: Google cloud storage folder.
        contains: Optional substring that must be in the file path.
        within_minutes: Optionally include only files that were updated in the
            last n minutes.
    """
    file_system = config["file_system"]()

    # (recursive doesn't work, so doing recursive search below)
    info = file_system.ls(folder, detail=True, recursive=True)

    if not info:
        return pd.DataFrame(columns=["kb", "mb", "name", "child", "path"])

    fileinfo = [
        (x["name"], x["size"], x["updated"])
        for x in info
        if x["storageClass"] != "DIRECTORY"
    ]
    folderinfo = [x["name"] for x in info if x["storageClass"] == "DIRECTORY"]

    fileinfo += _get_files_in_subfolders(folderinfo)

    df = pd.DataFrame(fileinfo, columns=["path", "kb", "updated"])

    if contains:
        try:
            df = df.loc[df["path"].str.contains(contains)]
        except TypeError:
            for item in contains:
                df = df.loc[df["path"].str.contains(item)]

    if not len(df):
        return df

    df = df.set_index("updated").sort_index()

    df["name"] = df["path"].apply(lambda x: Path(x).name)
    df["child"] = df["path"].apply(lambda x: Path(x).parent.name)

    df["kb"] = df["kb"].round(1)
    df["mb"] = (df["kb"] / 1_000_000).round(1)

    df.index = (
        pd.to_datetime(df.index)
        .round("s")
        .tz_convert("Europe/Oslo")
        .tz_localize(None)
        .round("s")
    )

    df.index.name = None

    if not within_minutes:
        return df[["kb", "mb", "name", "child", "path"]]

    the_time = pd.Timestamp.now() - pd.Timedelta(minutes=within_minutes)
    return df.loc[lambda x: x.index > the_time, ["kb", "mb", "name", "child", "path"]]


def _get_files_in_subfolders(folderinfo: list[dict]) -> list[tuple]:
    file_system = config["file_system"]()

    fileinfo = []

    while folderinfo:
        new_folderinfo = []
        for m in folderinfo:
            more_info = file_system.ls(m, detail=True, recursive=True)
            if not more_info:
                continue

            more_fileinfo = [
                (x["name"], x["size"], x["updated"])
                for x in more_info
                if x["storageClass"] != "DIRECTORY"
            ]
            fileinfo.extend(more_fileinfo)

            more_folderinfo = [
                x["name"]
                for x in more_info
                if x["storageClass"] == "DIRECTORY" and x["name"] not in folderinfo
            ]
            new_folderinfo.extend(more_folderinfo)

        folderinfo = new_folderinfo

    return fileinfo


def _maybe_strip_prefix(path, file_system):
    """Strip gcs prefix only if file_system is GCS."""
    if isinstance(file_system, GCSFileSystem) and path.startswith("gs://"):
        return path.replace("gs://", "")
    return path



import warnings

import joblib
import pandas as pd
import shapely
from geopandas import GeoDataFrame
from igraph import Graph
from pandas import DataFrame


def _get_route_frequencies(
    graph: Graph,
    roads: GeoDataFrame,
    weight_df: DataFrame,
    n_jobs: int,
) -> GeoDataFrame:
    """Function used in the get_route_frequencies method of NetworkAnalysis."""
    warnings.filterwarnings("ignore", category=RuntimeWarning)

    resultlist: list[DataFrame] = []

    od_pairs = weight_df.index

    ori_ids = od_pairs.get_level_values(0).unique()
    if n_jobs == 1:
        nested_results: list[list[DataFrame]] = [
            _get_one_route_frequency(
                ori_id, od_pairs=od_pairs, graph=graph, weight_df=weight_df
            )
            for ori_id in ori_ids
        ]
        del nested_results
    else:
        with joblib.Parallel(n_jobs) as parallel:
            nested_results: list[list[DataFrame]] = parallel(
                joblib.delayed(_get_one_route_frequency)(
                    ori_id, od_pairs=od_pairs, graph=graph, weight_df=weight_df
                )
                for ori_id in ori_ids
            )

    resultlist = [x for y in nested_results for x in y]

    if not resultlist:
        return pd.DataFrame(columns=["frequency", "geometry"])

    summarised: pd.Series = (
        pd.concat(resultlist).groupby("src_tgt_wt")["multiplier"].sum()
    )

    roads["frequency"] = roads["src_tgt_wt"].map(summarised)

    roads_visited = roads.loc[roads["frequency"].notna()].drop("src_tgt_wt", axis=1)

    return roads_visited


def _get_one_route_frequency(
    ori_id: int, od_pairs: pd.MultiIndex, graph: Graph, weight_df: pd.DataFrame
):
    relevant_pairs = od_pairs[od_pairs.get_level_values(0) == ori_id]
    destinations = relevant_pairs.get_level_values(1)

    res = graph.get_shortest_paths(
        weights="weight", v=ori_id, to=destinations, output="epath"
    )

    results = []
    for i, des_id in enumerate(destinations):
        indices = graph.es[res[i]]

        if not indices:
            continue

        line_ids = DataFrame({"src_tgt_wt": indices["src_tgt_wt"]})
        line_ids["origin"] = ori_id
        line_ids["destination"] = des_id
        line_ids["multiplier"] = weight_df.loc[ori_id, des_id].iloc[0]

        results.append(line_ids)
    return results


def _get_route(
    graph: Graph,
    weight: str,
    roads: GeoDataFrame,
    od_pairs: pd.MultiIndex,
    n_jobs: int,
) -> GeoDataFrame:
    """Function used in the get_route method of NetworkAnalysis."""
    warnings.filterwarnings("ignore", category=RuntimeWarning)

    ori_ids = od_pairs.get_level_values(0).unique()
    if n_jobs == 1:
        nested_results: list[list[DataFrame]] = [
            _get_one_route(ori_id, od_pairs=od_pairs, graph=graph) for ori_id in ori_ids
        ]
        del nested_results
    else:
        with joblib.Parallel(n_jobs) as parallel:
            nested_results: list[list[DataFrame]] = parallel(
                joblib.delayed(_get_one_route)(ori_id, od_pairs=od_pairs, graph=graph)
                for ori_id in ori_ids
            )

    resultlist = [x for y in nested_results for x in y]

    if not resultlist:
        warnings.warn(
            "No paths were found. Try larger search_tolerance or search_factor. "
            "Or close_network_holes() or remove_isolated().",
            stacklevel=1,
        )
        return pd.DataFrame(columns=["origin", "destination", weight, "geometry"])

    results: DataFrame = pd.concat(resultlist)
    assert list(results.columns) == ["origin", "destination"], list(results.columns)
    lines: GeoDataFrame = _get_line_geometries(results, roads, weight)
    lines.geometry = shapely.force_2d(lines.geometry)
    lines = lines.dissolve(by=["origin", "destination"], aggfunc="sum", as_index=False)
    return lines[["origin", "destination", weight, "geometry"]]


def _get_one_route(
    ori_id: int,
    od_pairs: pd.MultiIndex,
    graph: Graph,
) -> list[DataFrame]:
    relevant_pairs = od_pairs[od_pairs.get_level_values(0) == ori_id]
    destinations = relevant_pairs.get_level_values(1)

    results = graph.get_shortest_paths(
        weights="weight", v=ori_id, to=destinations, output="epath"
    )

    out_lines = []
    for i, des_id in enumerate(destinations):
        indices = graph.es[results[i]]
        if not indices:
            continue
        line_ids: DataFrame = _create_line_id_df(indices["src_tgt_wt"], ori_id, des_id)
        out_lines.append(line_ids)

    return out_lines


def _get_k_routes(
    graph: Graph,
    weight: str,
    roads: GeoDataFrame,
    k: int,
    drop_middle_percent: int,
    od_pairs: pd.MultiIndex,
) -> GeoDataFrame:
    """Function used in the get_k_routes method of NetworkAnalysis."""
    warnings.filterwarnings("ignore", category=RuntimeWarning)

    resultlist: list[DataFrame] = []

    for ori_id, des_id in od_pairs:
        k_lines: DataFrame = _loop_k_routes(
            graph, ori_id, des_id, k, drop_middle_percent
        )
        if k_lines is not None:
            resultlist.append(k_lines)

    if not resultlist:
        warnings.warn(
            "No paths were found. Try larger search_tolerance or search_factor. "
            "Or close_network_holes() or remove_isolated().",
            stacklevel=1,
        )
        return pd.DataFrame(columns=["origin", "destination", weight, "geometry"])

    results: DataFrame = pd.concat(resultlist)

    assert list(results.columns) == ["origin", "destination", "k"], list(
        results.columns
    )
    lines: GeoDataFrame = _get_line_geometries(results, roads, weight)

    lines = lines.dissolve(
        by=["origin", "destination", "k"], aggfunc="sum", as_index=False
    )

    return lines[["origin", "destination", weight, "k", "geometry"]]


def _loop_k_routes(
    graph: Graph,
    ori_id: str,
    des_id: str,
    k: int,
    drop_middle_percent: int,
) -> DataFrame:
    """Workaround for igraph's get_k_shortest_paths.

    igraph's get_k_shorest_paths doesn't seem to work (gives just the same path k
    times), so doing it manually. Run _get_one_route, then remove the edges in the
    middle of the route, given with drop_middle_percent, repeat k times.
    """
    graph = graph.copy()

    lines: list[DataFrame] = []

    for i in range(k):
        res = graph.get_shortest_paths(
            weights="weight", v=ori_id, to=des_id, output="epath"
        )
        if not res[0]:
            continue

        indices = graph.es[res[0]]

        line_ids = _create_line_id_df(indices["src_tgt_wt"], ori_id, des_id)
        line_ids["k"] = i + 1
        lines.append(line_ids)

        edge_tuples = indices["edge_tuples"]

        n_edges_to_keep = (
            len(edge_tuples) - len(edge_tuples) * drop_middle_percent / 100
        ) / 2

        n_edges_to_keep = int(round(n_edges_to_keep, 0))

        if n_edges_to_keep == 0:
            n_edges_to_keep = 1

        to_be_dropped = edge_tuples[n_edges_to_keep:-n_edges_to_keep]
        graph.delete_edges(to_be_dropped)

    if lines:
        return pd.concat(lines)
    else:
        return pd.DataFrame(columns=["origin", "destination", "k"])


def _get_line_geometries(
    line_ids: DataFrame, roads: GeoDataFrame, weight: str
) -> GeoDataFrame:
    road_mapper = roads.set_index(["src_tgt_wt"])[[weight, "geometry"]]
    line_ids = line_ids.join(road_mapper)
    return GeoDataFrame(line_ids, geometry="geometry", crs=roads.crs)


def _create_line_id_df(src_tgt_wt: list, ori_id: str, des_id: str) -> DataFrame:
    line_ids = DataFrame(index=src_tgt_wt)

    # remove edges from ori/des to the roads
    line_ids = line_ids.loc[~line_ids.index.str.endswith("_0")]

    line_ids["origin"] = ori_id
    line_ids["destination"] = des_id

    return line_ids



"""Make static maps with geopandas and matplotlib."""

import warnings
from typing import Any

import matplotlib
import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame

from ..geopandas_tools.conversion import to_bbox
from ..geopandas_tools.conversion import to_gdf
from ..helpers import is_property
from .legend import LEGEND_KWARGS
from .legend import ContinousLegend
from .legend import Legend
from .legend import prettify_bins
from .map import Map
from .map import _determine_best_name

# the geopandas._explore raises a deprication warning. Ignoring for now.
warnings.filterwarnings(
    action="ignore", category=matplotlib.MatplotlibDeprecationWarning
)
pd.options.mode.chained_assignment = None

MAP_KWARGS = {
    "bins",
    "title",
    "title_fontsize",
    "size",
    "cmap",
    "cmap_start",
    "cmap_stop",
    "scheme",
    "k",
    "column",
    "title_color",
    "facecolor",
    "labelcolor",
    "nan_color",
    # "alpha",
    "title_kwargs",
    "bg_gdf_color",
    "title_position",
    # "linewidth",
}


class ThematicMap(Map):
    """Class for making static maps.

    Args:
        *gdfs: One or more GeoDataFrames.
        column: The name of the column to plot.
        bounds: Optional bounding box for the map.
        title: Title of the plot.
        title_position: Title position. Either "center" (default), "left" or "right".
        size: Width and height of the plot in inches. Fontsize of title and legend is
            adjusted accordingly. Defaults to 25.
        dark: If False (default), the background will be white and the text black. If
            True, the background will be black and the text white. When True, the
            default cmap is "viridis", and when False, the default is red to purple
            (RdPu).
        cmap: Colormap of the plot. See:
            https://matplotlib.org/stable/tutorials/colors/colormaps.html
        scheme: How to devide numeric values into categories. Defaults to
            "naturalbreaks".
        k: Number of color groups.
        bins: For numeric columns. List of numbers that define the
            maximum value for the color groups.
        nan_label: Label for missing data.
        nan_hatch: Hatch for missing data. See https://matplotlib.org/stable/gallery/shapes_and_collections/hatch_style_reference.html.
        legend_kwargs: dictionary with attributes for the legend. E.g.:
            title: Legend title. Defaults to the column name.
            rounding: If positive number, it will round floats to n decimals.
            If negative, eg. -2, the number 3429 is rounded to 3400.
            By default, the rounding depends on the column's maximum value
            and standard deviation.
            position: The legend's x and y position in the plot. By default, it's
            decided dynamically by finding the space with most distance to
            the geometries. To be specified as a tuple of
            x and y position between 0 and 1. E.g. position=(0.8, 0.2) for a position
            in the bottom right corner, (0.2, 0.8) for the upper left corner.
            pretty_labels: Whether to capitalize words in text categories.
            label_suffix: For numeric columns. The text to put after each number
            in the legend labels. Defaults to None.
            label_sep: For numeric columns. Text to put in between the two numbers
            in each color group in the legend. Defaults to '-'.
            thousand_sep: For numeric columns. Separator between each thousand for
            large numbers. Defaults to None, meaning no separator.
            decimal_mark: For numeric columns. Text to use as decimal point.
            Defaults to None, meaning '.' (dot) unless 'thousand_sep' is
            '.'. In this case, ',' (comma) will be used as decimal mark.
        **kwargs: Additional attributes for the map. E.g.:
            title_color (str): Color of the title font.
            title_fontsize (int): Color of the title font.
            cmap_start (int): Start position for the color palette.
            cmap_stop (int): End position for the color palette.
            facecolor (str): Background color.
            labelcolor (str): Color for the labels.
            nan_color: Color for missing data.

    Examples:
    ---------
    >>> import sgis as sg
    >>> points = sg.random_points(100, loc=1000).pipe(sg.buff, np.random.rand(100) * 100)
    >>> points2 = sg.random_points(100, loc=1000).pipe(sg.buff, np.random.rand(100) * 100)


    Simple plot with legend and title.

    >>> m = sg.ThematicMap(points, points2, column="area", title="Area of random circles")
    >>> m.plot()

    Plot with custom legend units (label_suffix) and thousand separator.
    And with rounding set to -2, meaning e.g. 3429 is rounded to 3400.
    If rounding was set to positive 2, 3429 would be rounded to 3429.00.

    >>> m = sg.ThematicMap(
    ...     points,
    ...     points2,
    ...     column="area",
    ...     title = "Area of random circles",
    ...     legend_kwargs=dict(
    ...         rounding=-2,
    ...         thousand_sep=" ",
    ...         label_sep="to",
    ...     ),
    ... )
    >>> m.plot()

    With custom bins for the categories, and other customizations.

    >>> m = sg.ThematicMap(
    ...     points,
    ...     points2,
    ...     column="area",
    ...     cmap="Greens",
    ...     cmap_start=50,
    ...     cmap_stop=255,
    ...     nan_label="Missing",
    ...     title = "Area of random circles",
    ...     bins = [5000, 10000, 15000, 20000],
    ...     title_kwargs=dict(
    ...         loc="left",
    ...         y=0.93,
    ...         x=0.025,
    ...     ),
    ...     legend_kwargs=dict(
    ...         thousand_sep=" ",
    ...         label_sep="to",
    ...         decimal_mark=".",
    ...         label_suffix="m2",
    ...     ),
    ... )
    >>> m.plot()
    """

    def __init__(
        self,
        *gdfs: GeoDataFrame,
        column: str | None = None,
        bounds: tuple | None = None,
        title: str | None = None,
        title_position: tuple[float, float] | None = None,
        size: int = 25,
        dark: bool = False,
        cmap: str | None = None,
        scheme: str = "naturalbreaks",
        k: int = 5,
        bins: tuple[float] | None = None,
        nan_label: str = "Missing",
        nan_color: str | None = None,
        nan_hatch: str | None = None,
        hatch: str | None = None,
        legend_kwargs: dict | None = None,
        title_kwargs: dict | None = None,
        legend: bool = True,
        **kwargs,
    ) -> None:
        """Initializer."""
        new_gdfs = {}
        for i, gdf in enumerate(gdfs):
            if isinstance(gdf, str):
                raise ValueError("gdfs cannot be a string in ThematicMap.")
            name = _determine_best_name(gdf, column, i)
            if name in new_gdfs:
                name += str(i)
            try:
                new_gdfs[name] = to_gdf(gdf)
            except Exception:
                continue

        new_kwargs = {}
        self.kwargs = {}
        for key, value in kwargs.items():
            try:
                new_gdfs[key] = to_gdf(value)
            except Exception:
                new_kwargs[key] = value

        super().__init__(
            column=column,
            scheme=scheme,
            k=k,
            bins=bins,
            nan_label=nan_label,
            nan_color=nan_color,
            **new_gdfs,
        )

        self.kwargs = self.kwargs | new_kwargs

        self.title = title
        self._size = size
        self.hatch = hatch
        self._dark = dark
        self.title_kwargs = title_kwargs or {}
        if title_position and "position" in self.title_kwargs:
            raise TypeError(
                "Specify either 'title_position' or title_kwargs position, not both."
            )
        if title_position or "position" in self.title_kwargs:
            position = self.title_kwargs.pop("position", title_position)
            error_mess = (
                "legend_kwargs position should be a two length tuple/list with two numbers between "
                "0 and 1 (x, y position)"
            )
            if not hasattr(position, "__len__"):
                raise TypeError(error_mess)
            if len(position) != 2:
                raise ValueError(error_mess)
            x, y = position
            if "loc" not in self.title_kwargs:
                if x < 0.4:
                    self.title_kwargs["loc"] = "left"
                elif x > 0.6:
                    self.title_kwargs["loc"] = "right"
                else:
                    self.title_kwargs["loc"] = "center"

            self.title_kwargs["x"], self.title_kwargs["y"] = x, y
        self.background_gdfs = []

        legend_kwargs = legend_kwargs or {}

        self._title_fontsize = self._size * 1.9

        black = self.kwargs.pop("black", None)
        self._dark = self._dark or black

        if not self.cmap and not self._is_categorical:
            self._choose_cmap(**kwargs)

        if not legend:
            self.legend = None
        else:
            self._create_legend()

        new_kwargs = {}
        for key, value in self.kwargs.items():
            if key not in MAP_KWARGS:
                new_kwargs[key] = value
            elif is_property(self, key):
                setattr(self, f"_{key}", value)
            else:
                setattr(self, key, value)
        self.kwargs = new_kwargs

        for key, value in legend_kwargs.items():
            if key not in LEGEND_KWARGS:
                raise TypeError(
                    f"{self.__class__.__name__} legend_kwargs got an unexpected key {key}"
                )
            if self.legend is not None:
                try:
                    setattr(self.legend, key, value)
                except Exception:
                    setattr(self.legend, f"_{key}", value)

        self._dark_or_light()

        if cmap:
            self._cmap = cmap

        self.bounds = (
            to_bbox(bounds) if bounds is not None else to_bbox(self._gdf.total_bounds)
        )
        self.minx, self.miny, self.maxx, self.maxy = self.bounds
        self.diffx = self.maxx - self.minx
        self.diffy = self.maxy - self.miny

        if self._gdf[self._column].isna().any():
            isnas = []
            for label, gdf in self._gdfs.items():
                isnas.append(gdf[gdf[self._column].isna()])
                self._gdfs[label] = gdf[gdf[self._column].notna()]
            self._more_data[nan_label] = {
                "gdf": pd.concat(isnas, ignore_index=True),
                "color": self.nan_color,
                "hatch": nan_hatch,
            } | new_kwargs
            self._gdf = pd.concat(self.gdfs.values(), ignore_index=True)

    @property
    def valid_keywords(self) -> set[str]:
        """List all valid keywords for the class initialiser."""
        return MAP_KWARGS

    def change_cmap(self, cmap: str, start: int = 0, stop: int = 256) -> "ThematicMap":
        """Change the color palette of the plot.

        Args:
            cmap: The colormap.
                https://matplotlib.org/stable/tutorials/colors/colormaps.html
            start: Start position for the color palette. Defaults to 0.
            stop: End position for the color palette. Defaults to 256, which
                is the end of the color range.
        """
        super().change_cmap(cmap, start, stop)
        return self

    def add_data(
        self,
        *,
        color: str | None = None,
        hatch: str | None = None,
        **kwargs,
    ) -> "ThematicMap":
        """Add Geometric Data of a given color or hatch.

        The geodata must be passed as keyword argument.
        The keyword will be used as label in the legend.

        Args:
            color: Color of the data.
            hatch: Hatch of the data. See https://matplotlib.org/stable/gallery/shapes_and_collections/hatch_style_reference.html.
            **kwargs: Must include exactly one GeoDataFrame or object that can be converted to GeoDataFrame.
                Additional kwargs are passed to geopandas.GeoDataFrame.plot and matplotlib.patches.Patch.
        """
        new_kwargs = {}
        n = 0
        for key, value in kwargs.items():
            try:
                gdf = to_gdf(value)
                label = key
                n += 1
            except Exception:
                new_kwargs[key] = value

        if n != 1:
            raise ValueError(
                "Must specify a single geometry object as keyword argument."
            )
        if not color and not hatch:
            raise TypeError("Must pass either 'color' or 'hatch'.")
        if hatch is not None:
            color = self.facecolor
        gdf["label"] = label
        self._more_data[label] = {
            "gdf": gdf,
            "color": color,
            "hatch": hatch,
        } | new_kwargs
        if self.bounds is None:
            self.bounds = to_bbox(self._gdf.total_bounds)
        return self

    def add_background(
        self,
        gdf: GeoDataFrame,
        color: str | None = None,
        **kwargs,
    ) -> "ThematicMap":
        """Add a GeoDataFrame as a background layer.

        Args:
            gdf: a GeoDataFrame.
            color: Single color. Defaults to gray (shade depends on whether the map
                facecolor is black or white).
            **kwargs: Keyword arguments sent to GeoDataFrame.plot.
        """
        if color:
            self.bg_gdf_color = color
        if not hasattr(self, "_background_gdfs"):
            self._background_gdfs = gdf
        else:
            self._background_gdfs = pd.concat(
                [self._background_gdfs, gdf], ignore_index=True
            )
        if self.bounds is None:
            self.bounds = to_bbox(self._gdf.total_bounds)
        self.bg_gdf_kwargs = kwargs
        return self

    def plot(self, **kwargs) -> None:
        """Creates the final plot.

        This method should be run after customising the map, but before saving.

        """
        kwargs = kwargs | self.kwargs
        __test = kwargs.pop("__test", False)
        include_legend = bool(kwargs.pop("legend", self.legend))

        if "color" in kwargs:
            kwargs.pop("column", None)
            self.legend = None
            include_legend = False
        elif hasattr(self, "color"):
            kwargs.pop("column", None)
            kwargs["color"] = self.color
            self.legend = None
            include_legend = False

        elif self._is_categorical:
            self._prepare_categorical_plot()
            if self._gdf is not None:
                kwargs["color"] = self._gdf["color"]
            if self.legend:
                self.legend._prepare_categorical_legend(
                    categories_colors=self._categories_colors_dict,
                    hatch=self.hatch,
                )

        else:
            kwargs = self._prepare_continous_plot(kwargs)
            if self.legend:
                if not self.legend.rounding:
                    self.legend._rounding = self.legend._get_rounding(
                        array=self._gdf[self._column].dropna()
                    )

                self.legend._prepare_continous_legend(
                    bins=self.bins,
                    colors=self._unique_colors,
                    bin_values=self._bins_unique_values,
                    nan_label=self.nan_label,
                    hatch=self.hatch,
                )

        if self.legend and self._more_data:
            self.legend._add_more_data_to_legend(self._more_data)

        if self.legend and not self.legend._position_has_been_set:
            self.legend._position = self.legend._get_best_legend_position(
                self._gdf, k=self._k + self._gdf[self._column].isna().any()
            )

        self._prepare_plot(**kwargs)

        if self.legend:
            self.ax = self.legend._actually_add_legend(ax=self.ax)

        self.ax = self._gdf.plot(
            legend=include_legend, ax=self.ax, hatch=self.hatch, **kwargs
        )

        if __test:
            return self

    def save(self, path: str) -> None:
        """Save figure as image file.

        To be run after the plot method.

        Args:
            path: File path.
        """
        try:
            plt.savefig(path)
        except FileNotFoundError:
            from dapla import FileClient

            fs = FileClient.get_gcs_file_system()
            with fs.open(path, "wb") as file:
                plt.savefig(file)

    def _prepare_plot(self, **kwargs) -> None:
        """Add figure and axis, title and background gdf."""
        for attr in self.__dict__.keys():
            if attr in self.kwargs:
                self[attr] = self.kwargs.pop(attr)
            if attr in kwargs:
                self[attr] = kwargs.pop(attr)

        self.fig, self.ax = self._get_matplotlib_figure_and_axix(
            figsize=(self._size, self._size)
        )
        self.fig.patch.set_facecolor(self.facecolor)
        self.ax.set_axis_off()
        self.ax.set_xlim([self.minx - self.diffx * 0.03, self.maxx + self.diffx * 0.03])
        self.ax.set_ylim([self.miny - self.diffy * 0.03, self.maxy + self.diffy * 0.03])

        if hasattr(self, "_background_gdfs"):
            self._actually_add_background()
        # if self.bounds is not None:

        for datadict in self._more_data.values():
            gdf = datadict["gdf"]
            datadict = {key: value for key, value in datadict.items() if key != "gdf"}
            gdf.plot(ax=self.ax, **datadict)

        if self.title:
            self.ax.set_title(
                self.title,
                **(
                    dict(fontsize=self.title_fontsize, color=self.title_color)
                    | self.title_kwargs
                ),
            )

    def _prepare_continous_plot(self, kwargs: dict) -> dict:
        """Create bins and colors."""
        self._prepare_continous_map()

        if self.scheme is None:
            self.legend = None
            kwargs["column"] = self.column
            return kwargs

        elif self.bins is None:
            kwargs["column"] = self.column
            return kwargs

        else:
            if self.legend and self.legend.rounding and self.legend.rounding < 0:
                self.bins = prettify_bins(self.bins, self.legend.rounding)
                self.bins = list({round(bin_, 5) for bin_ in self.bins})
                self.bins.sort()
                # self.legend._rounding_was = self.legend.rounding
                # self.legend.rounding = None

            classified = self._classify_from_bins(self._gdf, bins=self.bins)
            classified_sequential = self._push_classification(classified)
            n_colors = (
                len(np.unique(classified_sequential))
                - self._gdf[self._column].isna().any()
            )
            self._unique_colors = self._get_continous_colors(n=n_colors)
            self._bins_unique_values = self._make_bin_value_dict(
                self._gdf, classified_sequential
            )

            colorarray = self._unique_colors[classified_sequential]
            kwargs["color"] = colorarray

        if (
            self.legend and self.legend.rounding
        ):  # not self.legend._rounding_has_been_set:
            self.bins = self.legend._set_rounding(
                bins=self.bins, rounding=self.legend._rounding
            )

            if self._gdf[self._column].isna().any():
                self.bins = self.bins + [self.nan_label]

        return kwargs

    def _actually_add_legend(self) -> None:
        """Add legend to the axis and fill it with colors and labels."""
        if not self.legend._position_has_been_set:
            self.legend._position = self.legend._get_best_legend_position(
                self._gdf, k=self._k + self._gdf[self._column].isna().any()
            )

        if self._is_categorical:
            self.ax = self.legend._actually_add_categorical_legend(
                ax=self.ax,
                categories_colors=self._categories_colors_dict,
                nan_label=self.nan_label,
            )
        else:
            self.ax = self.legend._actually_add_continous_legend(
                ax=self.ax,
                bins=self.bins,
                colors=self._unique_colors,
                nan_label=self.nan_label,
                bin_values=self._bins_unique_values,
            )

    def _create_legend(self) -> None:
        """Instantiate the Legend class."""
        if self._is_categorical:
            self.legend = Legend(title=self._column, size=self._size)
        else:
            self.legend = ContinousLegend(title=self._column, size=self._size)

    def _choose_cmap(self, **kwargs) -> None:
        """Kwargs is to catch start and stop points for the cmap in __init__."""
        if self._dark:
            self._cmap = "viridis"
            self.cmap_start = self.kwargs.pop("cmap_start", 0)
            self.cmap_stop = self.kwargs.pop("cmap_stop", 256)
        else:
            self._cmap = "RdPu"
            self.cmap_start = self.kwargs.pop("cmap_start", 23)
            self.cmap_stop = self.kwargs.pop("cmap_stop", 256)

    def _make_bin_value_dict(self, gdf: GeoDataFrame, classified: np.ndarray) -> dict:
        """Dict with unique values of all bins. Used in labels in ContinousLegend."""
        bins_unique_values = {
            i: list(set(gdf.loc[classified == i, self._column]))
            for i, _ in enumerate(np.unique(classified))
        }
        return bins_unique_values

    def _actually_add_background(self) -> None:
        self._background_gdfs.plot(
            ax=self.ax, color=self.bg_gdf_color, **self.bg_gdf_kwargs
        )

    @staticmethod
    def _get_matplotlib_figure_and_axix(
        figsize: tuple[int, int],
    ) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(1, 1, 1)
        return fig, ax

    def _dark_or_light(self) -> None:
        if self._dark:
            self.facecolor, self.title_color, self.bg_gdf_color = (
                "#0f0f0f",
                "#fefefe",
                "#383836",
            )
            self.nan_color = "#666666" if self._nan_color_was_none else self.nan_color
            if not self._is_categorical:
                self._cmap = self.kwargs.get("cmap", "viridis")

            if self.legend is not None:
                for key, color in {
                    "facecolor": "#0f0f0f",
                    "labelcolor": "#fefefe",
                    "title_color": "#fefefe",
                    "edgecolor": "#0f0f0f",
                }.items():
                    setattr(self.legend, key, color)

        else:
            self.facecolor, self.title_color, self.bg_gdf_color = (
                "#fefefe",
                "#0f0f0f",
                "#e8e6e6",
            )
            self.nan_color = "#c2c2c2" if self._nan_color_was_none else self.nan_color
            if not self._is_categorical:
                self._cmap = self.kwargs.get("cmap", "RdPu")

            if self.legend is not None:
                for key, color in {
                    "facecolor": "#fefefe",
                    "labelcolor": "#0f0f0f",
                    "title_color": "#0f0f0f",
                    "edgecolor": "#0f0f0f",
                }.items():
                    setattr(self.legend, key, color)

    @property
    def title_fontsize(self) -> int:
        """Title fontsize, not to be confused with legend.title_fontsize."""
        return self._title_fontsize

    @title_fontsize.setter
    def title_fontsize(self, new_value: int) -> None:
        self._title_fontsize = new_value
        self._title_fontsize_has_been_set = True

    @property
    def size(self) -> int:
        """Size of the image."""
        return self._size

    @size.setter
    def size(self, new_value: bool) -> None:
        """Adjust font and marker size if not actively set."""
        self._size = new_value
        if not hasattr(self, "_title_fontsize_has_been_set"):
            self._title_fontsize = self._size * 2
        if not hasattr(self, "legend"):
            return
        if not hasattr(self.legend, "_title_fontsize_has_been_set"):
            self.legend._title_fontsize = self._size * 1.2
        if not hasattr(self.legend, "_fontsize_has_been_set"):
            self.legend._fontsize = self._size
        if not hasattr(self.legend, "_markersize_has_been_set"):
            self.legend._markersize = self._size

    def __setattr__(self, __name: str, __value: Any) -> None:
        """Set an attribute with square brackets."""
        if "legend_" in __name:
            last_part = __name.split("legend_")[-1]
            raise AttributeError(
                f"Invalid attribute {__name!r}. Did you mean 'legend.{last_part}'?"
            )
        return super().__setattr__(__name, __value)
